# Copyright (c) 2017 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""Common submodule containing configuration parsing and backend utilities."""



# Copyright (c) 2018 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""The signac framework aids in the management of large and heterogeneous data spaces.

It provides a simple and robust data model to create a well-defined, indexable
storage layout for data and metadata. This makes it easier to operate on large
data spaces, streamlines post-processing and analysis, and makes data
collectively accessible.
"""

from . import cite, contrib, db, errors, sync, testing, warnings
from .contrib import (
    Collection,
    MainCrawler,
    MasterCrawler,
    Project,
    RegexFileCrawler,
    SignacProjectCrawler,
    TemporaryProject,
    export,
    export_one,
    export_pymongo,
    export_to_mirror,
    fetch,
)
from .contrib import filesystems as fs
from .contrib import get_job, get_project, index, index_files, init_project
from .core.h5store import H5Store, H5StoreManager
from .core.jsondict import JSONDict
from .core.jsondict import buffer_reads_writes as buffered
from .core.jsondict import flush_all as flush
from .core.jsondict import get_buffer_load, get_buffer_size
from .core.jsondict import in_buffered_mode as is_buffered
from .db import get_database
from .diff import diff_jobs
from .version import __version__

__all__ = [
    "__version__",
    "contrib",
    "db",
    "errors",
    "warnings",
    "sync",
    "cite",
    "Project",
    "TemporaryProject",
    "get_project",
    "init_project",
    "get_job",
    "diff_jobs",
    "get_database",
    "fetch",
    "export_one",
    "export",
    "export_to_mirror",
    "Collection",
    "export_pymongo",
    "fs",
    "index_files",
    "index",
    "RegexFileCrawler",
    "MainCrawler",
    "MasterCrawler",
    "SignacProjectCrawler",
    "buffered",
    "is_buffered",
    "flush",
    "get_buffer_size",
    "get_buffer_load",
    "JSONDict",
    "H5Store",
    "H5StoreManager",
    "testing",
]



# Copyright (c) 2017 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""Errors raised by signac.core classes."""


class Error(Exception):
    """Base class used for signac Errors."""

    pass



# Copyright (c) 2017 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""Defines attribute dictionaries, where values can be accessed as attributes."""

from .synceddict import _SyncedDict


class SyncedAttrDict(_SyncedDict):
    """A synced dictionary where (nested) values can be accessed as attributes.

    For example:

    .. code-block:: python

        nested_dict = dict(a=dict(b=0))
        ad = SyncedAttrDict(nested_dict)
        assert ad.a.b == 0
    """

    _PROTECTED_KEYS = ("_data", "_suspend_sync_", "_load", "_save")

    def __getattr__(self, name):
        try:
            return super().__getattribute__(name)
        except AttributeError:
            if name.startswith("__"):
                raise
            try:
                return self.__getitem__(name)
            except KeyError as e:
                raise AttributeError(e)

    def __setattr__(self, key, value):
        try:
            super().__getattribute__("_data")
        except AttributeError:
            super().__setattr__(key, value)
        else:
            if key.startswith("__") or key in self.__getattribute__("_PROTECTED_KEYS"):
                super().__setattr__(key, value)
            else:
                self.__setitem__(key, value)

    def __delattr__(self, key):
        if key.startswith("__") or key in self._PROTECTED_KEYS:
            super().__delattr__(key)
        else:
            self.__delitem__(key)



# Copyright (c) 2017 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""Errors raised by signac.common classes."""
from ..core.errors import Error


class ConfigError(Error, RuntimeError):
    """Error with parsing or reading a configuration file."""

    pass


# this class is only used by deprecated features
class AuthenticationError(Error, RuntimeError):
    """Authentication error."""

    def __str__(self):
        if len(self.args) > 0:
            return "Failed to authenticate with host '{}'.".format(self.args[0])
        else:
            return "Failed to authenticate with host."


# this class is only used by deprecated features
class ExportError(Error, RuntimeError):
    """Error exporting documents to a mirror."""

    pass


# this class is only used by deprecated features
class FetchError(FileNotFoundError):
    """Error in fetching data."""

    pass



# Copyright (c) 2017 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""Utility classes for signac version."""

import re
import subprocess

from deprecation import deprecated

from ..version import __version__


@deprecated(
    deprecated_in="1.3",
    removed_in="2.0",
    current_version=__version__,
    details="All database related functions have been deprecated.",
)
def get_subject_from_certificate(fn_certificate):  # noqa: D103, E261
    try:
        cert_txt = subprocess.check_output(
            [
                "openssl",
                "x509",
                "-in",
                fn_certificate,
                "-inform",
                "PEM",
                "-subject",
                "-nameopt",
                "RFC2253",
            ]
        ).decode()
    except subprocess.CalledProcessError:
        msg = "Unable to retrieve subject from certificate '{}'."
        raise RuntimeError(msg.format(fn_certificate))
    else:
        lines = cert_txt.split("\n")
        assert lines[0].startswith("subject=")
        return lines[0][len("subject=") :].strip()


class Version(dict):
    """Utility class to manage revision control numbers."""

    def __init__(self, major=0, minor=0, change=0, postrelease="", prerelease="final"):
        if prerelease > "final":
            raise ValueError("illegal pre-release tag", prerelease)
        super().__init__(
            major=major,
            minor=minor,
            change=change,
            postrelease=postrelease,
            prerelease=prerelease,
        )

    def to_tuple(self):
        """Return version details as tuple."""
        return (
            self["major"],
            self["minor"],
            self["change"],
            self["prerelease"],
            self["postrelease"],
        )

    def __lt__(self, other):
        return self.to_tuple() < other.to_tuple()

    def __eq__(self, other):
        return self.to_tuple() == other.to_tuple()

    def __str__(self):
        return "{major}.{minor}{postrelease}.{change}{prerelease}".format(**self)

    def __repr__(self):
        return "Version({})".format(",".join((f"{k}={v}" for k, v in self.items())))


def parse_version(version_str):
    """Parse a version number into a version object."""
    p = re.compile(
        r"(?P<major>[0-9]*)\.(?P<minor>[0-9]*)((?P<postrelease>-?\w*)"
        r"\.(?P<change>[0-9])(?P<prerelease>\w*))?"
    )
    r = p.match(version_str)
    v = r.groupdict()
    version = Version(
        **{
            "major": int(v.get("major") or 0),
            "minor": int(v.get("minor") or 0),
            "change": int(v.get("change") or 0),
            "postrelease": str(v.get("postrelease") or ""),
            "prerelease": str(v.get("prerelease") or "final"),
        }
    )
    return version



# Copyright (c) 2017 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
from deprecation import deprecated

from ..common import host
from ..version import __version__

"""
THIS MODULE IS DEPRECATED!
"""


@deprecated(
    deprecated_in="1.3",
    removed_in="2.0",
    current_version=__version__,
    details="The database package is deprecated.",
)
def get_database(name, hostname=None, config=None):
    """Get a database handle.

    The database handle is an instance of :class:`~pymongo.database.Database`,
    which provides access to the document collections within one database.

    .. code-block:: python

        db = signac.db.get_database('MyDatabase')
        docs = db.my_collection.find()

    Please note, that a collection which did not exist at the point of access,
    will automatically be created.

    :param name: The name of the database to get.
    :type name: str
    :param hostname: The name of the configured host.
                     Defaults to the first configured host, or the
                     host specified by `default_host`.
    :type hostname: str
    :param config: The config object to retrieve the host
                   configuration from.
                   Defaults to the global configuration.
    :type config: :class:`.common.config.Config`
    :returns: The database handle.
    :rtype: :class:`pymongo.database.Database`

    .. seealso:: https://api.mongodb.org/python/current/api/pymongo/database.html
    """
    return host.get_database(name=name, hostname=hostname, config=config)



# Copyright (c) 2019 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""Functions for initializing workspaces for testing purposes."""
from itertools import cycle


def init_jobs(project, nested=False, listed=False, heterogeneous=False):
    """Initialize a dataspace for testing purposes.

    Parameters
    ----------
    project : :class:`~signac.Project`
        The project where jobs will be added.
    nested : bool
        If True, included nested state points (Default value = False).
    listed : bool
        If True, include lists as values of state point parameters (Default
        value = False).
    heterogeneous : bool
        If True, include heterogeneous state point parameters (Default value =
        False).

    Returns
    -------
    list[:class:`~signac.contrib.job.Job`]
        A list containing the initialized jobs.

    """
    jobs_init = []
    vals = [1, 1.0, "1", False, True, None]
    if nested:
        vals += [{"b": v, "c": 0} if heterogeneous else {"b": v} for v in vals]
    if listed:
        vals += [[v, 0] if heterogeneous else [v] for v in vals]
    if heterogeneous:
        for k, v in zip(cycle("ab"), vals):
            jobs_init.append(project.open_job({k: v}).init())
    else:
        for v in vals:
            jobs_init.append(project.open_job(dict(a=v)).init())
    return jobs_init



# Copyright (c) 2017 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""Contrib submodule containing Project class and indexing features."""

import logging

from . import indexing
from .collection import Collection
from .indexing import (
    BaseCrawler,
    JSONCrawler,
    MainCrawler,
    MasterCrawler,
    RegexFileCrawler,
    SignacProjectCrawler,
    export,
    export_one,
    export_pymongo,
    export_to_mirror,
    fetch,
    fetched,
    index,
    index_files,
)
from .project import Project, TemporaryProject, get_job, get_project, init_project

logger = logging.getLogger(__name__)


__all__ = [
    "indexing",
    "Project",
    "TemporaryProject",
    "get_project",
    "init_project",
    "get_job",
    "BaseCrawler",
    "RegexFileCrawler",
    "JSONCrawler",
    "SignacProjectCrawler",
    "MainCrawler",
    "MasterCrawler",
    "fetch",
    "fetched",
    "export_one",
    "export",
    "export_to_mirror",
    "export_pymongo",
    "index_files",
    "index",
    "Collection",
]


try:
    import mpi4py  # noqa
except ImportError:
    logger.debug("Failed to import mpi4py. MPIPool will not be available.")
else:
    from .mpipool import MPIPool  # noqa

    __all__.extend(["MPIPool"])



# Copyright (c) 2017 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
import logging
import warnings

try:
    import pymongo  # noqa
except ImportError:
    warnings.warn(
        "Failed to import pymongo. " "get_database will not be available.",
        ImportWarning,
    )

    def get_database(*args, **kwargs):
        """Get a database handle.

        This function is only available if pymongo is installed."""
        raise ImportError("You need to install pymongo to use `get_database()`.")


else:
    if pymongo.version_tuple[0] < 3:
        logging.getLogger(__name__).warn(
            "Your pymongo installation (version {}) is no longer "
            "supported by signac. Consider updating.".format(pymongo.version)
        )
    from .database import get_database


__all__ = ["get_database"]



# Copyright (c) 2018 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""Linked view classes."""

import errno
import logging
import os
import sys
from itertools import chain

from .utility import _mkdir_p

logger = logging.getLogger(__name__)


def create_linked_view(project, prefix=None, job_ids=None, index=None, path=None):
    """Create or update a persistent linked view of the selected data space.

    Parameters
    ----------
    project : signac.Project
        Project handle.
    prefix : str
        The path where the linked view will be created or updated (Default value = None).
    job_ids : iterable
        If None (the default), create the view for the complete data space,
        otherwise only for this iterable of job ids.
    index :
        A document index (Default value = None).
    path :
        The path (function) used to structure the linked data space (Default value = None).

    Returns
    -------
    dict
        A dictionary that maps the source directory paths to the linked
        directory paths.

    Raises
    ------
    OSError
        Linked views cannot be created on Windows because
        symbolic links are not supported by the platform.
    ValueError
        When the selected data space is provided with an insufficient index.
    RuntimeError
        When state points contain one of ``[os.sep, " ", "*"]``.

    """
    from .import_export import _check_directory_structure_validity, _make_path_function

    # Windows does not support the creation of symbolic links.
    if sys.platform == "win32":
        raise OSError(
            "signac cannot create linked views on Windows, because "
            "symbolic links are not supported by the platform."
        )

    if prefix is None:
        prefix = "view"

    if index is None:
        if job_ids is None:
            index = [{"_id": job._id, "statepoint": job.sp()} for job in project]
            jobs = list(project)
        else:
            index = [
                {"_id": job_id, "statepoint": project.open_job(id=job_id).sp()}
                for job_id in job_ids
            ]
            jobs = list(project.open_job(id=job_id) for job_id in job_ids)
    elif job_ids is not None:
        if not isinstance(job_ids, set):
            job_ids = set(job_ids)
        index = [doc for doc in index if doc["_id"] in job_ids]
        jobs = list(project.open_job(id=job_id) for job_id in job_ids)
        if not job_ids.issubset({doc["_id"] for doc in index}):
            raise ValueError("Insufficient index for selected data space.")

    key_list = [k for job in jobs for k in job.statepoint().keys()]
    value_list = [v for job in jobs for v in job.statepoint().values()]
    item_list = key_list + value_list
    bad_chars = [os.sep, " ", "*"]
    bad_items = [
        item
        for item in item_list
        for char in bad_chars
        if isinstance(item, str) and char in item
    ]

    if any(bad_items):
        err_msg = " ".join(
            [
                f"In order to use view, statepoints should not contain {bad_chars}:",
                *bad_items,
            ]
        )
        raise RuntimeError(err_msg)

    path_function = _make_path_function(jobs, path)

    links = {}
    for job in jobs:
        paths = os.path.join(path_function(job), "job")
        links[paths] = job.workspace()
    if not links:  # data space contains less than two elements
        for job in project.find_jobs():
            links["./job"] = job.workspace()
        assert len(links) < 2
    _check_directory_structure_validity(links.keys())

    _update_view(prefix, links)
    return links


def _update_view(prefix, links, leaf="job"):
    """Update an existing linked view hierarchy in place.

    Parameters
    ----------
    prefix : str
        The path where the linked view will be created or updated (Default value = None).
    links : dict
        Linked view .
    leaf : str
        The name of the leaf directories in the view
        directory tree (Default value = 'job').

    """
    obsolete, to_update, new = _analyze_view(prefix, links)
    num_ops = len(obsolete) + 2 * len(to_update) + len(new)
    if num_ops:
        logger.info(f"Generating current view in '{prefix}' ({num_ops} operations)...")
    else:
        logger.info(f"View in '{prefix}' is up to date.")
        return
    logger.debug("Removing {} obsolete links.".format(len(obsolete)))
    for path in obsolete:
        p = os.path.join(prefix, path)
        try:
            os.unlink(p)
        except OSError:
            os.rmdir(p)
    logger.debug(
        "Creating {} new and updating {} existing links.".format(
            len(new), len(to_update)
        )
    )
    for path in to_update:
        os.unlink(os.path.join(prefix, path))
    for path in chain(new, to_update):
        dst = os.path.join(prefix, path)
        src = os.path.relpath(links[path], os.path.split(dst)[0])
        _make_link(src, dst)


def _analyze_view(prefix, links, leaf="job"):
    """Analyze an existing view to prepare for update.

    Parameters
    ----------
    prefix : str
        The path where the linked view will be created or updated (Default value = None).
    links : dict
        Linked view.
    leaf : str
        The name of the leaf directories in the view
        directory tree (Default value = 'job').

    Returns
    -------
    tuple
        tuple that contains: (list of outdated links, list of links to update, set of new links).

    """
    logger.info(f"Analyzing view prefix '{prefix}'...")
    existing_paths = {os.path.join(p, leaf) for p in _find_all_links(prefix, leaf)}
    existing_tree = _build_tree(existing_paths)
    for path in links:
        _color_path(existing_tree, path.split(os.sep))
    obsolete = []
    dead_branches = _find_dead_branches(existing_tree)
    for branch in reversed(sorted(dead_branches, key=len)):
        if branch:
            obsolete.append(os.path.join(*(n.name for n in branch)))
    if "." in obsolete:
        obsolete.remove(".")
    keep_or_update = existing_paths.intersection(links.keys())
    new = set(links.keys()).difference(keep_or_update)
    to_update = [
        p
        for p in keep_or_update
        if os.path.realpath(os.path.join(prefix, p)) != links[p]
    ]
    return obsolete, to_update, new


def _make_link(src, dst):
    """Create a symbolic link and all directories leading to it.

    Parameters
    ----------
    src : str
        Name of directory/file to create a symbolic link.
    dst : str
        Destination symbolic link directory/file name.

    """
    _mkdir_p(os.path.dirname(dst))
    try:
        os.symlink(src, dst, target_is_directory=True)
    except OSError as error:
        if error.errno == errno.EEXIST:
            if os.path.realpath(src) == os.path.realpath(dst):
                return
        raise


def _find_all_links(root, leaf="job"):
    """Find all symbolic links under root.

    Parameters
    ----------
    root : str
        Project root directory.
    leaf : str
        The name of the leaf directories in the view
        directory tree (Default value = 'job').

    Yields
    ------
    str
        Relative path to root from ``leaf``.

    """
    for dirpath, dirnames, filenames in os.walk(root):
        for dirname in dirnames:
            if dirname == leaf:
                yield os.path.relpath(dirpath, root)
                break
        for filename in filenames:
            if filename == leaf:
                yield os.path.relpath(dirpath, root)
                break


class _Node:
    """Generic graph-node class."""

    def __init__(self, name=None, value=None):
        self.name = name
        self.value = value
        self.children = {}

    def get_child(self, name):
        """Get child node corresponding to the name passed.

        Parameters
        ----------
        name : str
            Name of child node to get.

        Returns
        -------
        :class:`~signac.contrib.linked_view._Node`
            The requested child node.

        """
        return self.children.setdefault(name, type(self)(name))

    def __str__(self):
        return f"_Node({self.name}, {self.value})"

    __repr__ = __str__


def _build_tree(paths):
    """Build a graph structure for paths.

    Parameters
    ----------
    paths : list
        A list of paths to views that already exist.

    Returns
    -------
    :class:`~signac.contrib.linked_view._Node`
        Graph structure for path.

    """
    root = _Node()
    for path in paths:
        node = root
        for p in path.split(os.sep):
            node = node.get_child(p)
    return root


def _get_branches(root, branch=None):
    """Get all branches from the root node.

    Parameters
    ----------
    root : :class:`~signac.contrib.linked_view._Node`
        Root node.
    branch : list
        The current list of branches that has been collected,
        used in recursive calls to build up the branches starting
        at the root (Default value = None).

    Yields
    ------
    list
        Branches for the root node.

    """
    if branch is None:
        branch = []
    else:
        branch = list(branch) + [root]
    if root.children:
        for child in root.children.values():
            yield from _get_branches(child, branch)
    else:
        yield branch


def _color_path(root, path):
    """Color the path from root by setting value to True.

    Parameters
    ----------
    root : :class:`~signac.contrib.linked_view._Node`
        Root node.
    path : list
        The name of the directory/file to color (set value to True).

    """
    root.value = True
    for name in path:
        root = root.get_child(name)
        root.value = True


def _find_dead_branches(root, branch=None):
    """Find all branches considered dead (not-colored).

    Parameters
    ----------
    root : :class:`~signac.contrib.linked_view._Node`
        Root node.
    branch : list
        The current list of branches that has been collected,
        used in recursive calls to build up the branches starting
        at the root (Default value = None).

    Yields
    ------
    list
        Branches that are considered as dead (not-colored).

    """
    if branch is None:
        branch = []
    else:
        branch = list(branch) + [root]
    if root.children:
        for child in root.children.values():
            yield from _find_dead_branches(child, branch)
    if not root.value:
        yield branch



# Copyright (c) 2017 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""Functions to support citing this software."""
import sys

ARXIV_BIBTEX = """@online{signac,
    author      = {Carl S. Adorf and Paul M. Dodd and Sharon C. Glotzer},
    title       = {signac - A Simple Data Management Framework},
    year        = {2016},
    eprinttype  = {arxiv},
    eprintclass = {cs.DB},
    eprint      = {1611.03543}
}
"""


ARXIV_REFERENCE = (
    "Carl S. Adorf, Paul M. Dodd, and Sharon C. Glotzer. "
    "signac - A Simple Data Management Framework. 2016. "
    "arXiv:1611.03543 [cs.DB]"
)


def bibtex(file=None):
    """Generate bibtex entries for signac.

    The bibtex entries will be printed to screen unless a
    filename or a file-like object are provided, in which
    case they will be written to the corresponding file.

    .. note::

        A full reference should also include the
        version of this software. Please refer to the
        documentation on how to cite a specific version.

    :param file: A str or file-like object.
        Defaults to sys.stdout.
    """
    if file is None:
        file = sys.stdout
    elif isinstance(file, str):
        file = open(file, "w")
    file.write(ARXIV_BIBTEX)


def reference(file=None):
    """Generate formatted reference entries for signac.

    The references will be printed to screen unless a
    filename or a file-like object are provided, in which
    case they will be written to the corresponding file.

    .. note::

        A full reference should also include the
        version of this software. Please refer to the
        documentation on how to cite a specific version.

    :param file: A str or file-like object.
        Defaults to sys.stdout.
    """
    if file is None:
        file = sys.stdout
    elif isinstance(file, str):
        file = open(file, "w")
    file.write(ARXIV_REFERENCE + "\n")



# Copyright (c) 2017 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""Errors raised by signac.contrib classes."""

from ..core.errors import Error


class WorkspaceError(Error, OSError):
    """Raised when there is an issue creating or accessing the workspace.

    Parameters
    ----------
    error :
        The underlying error causing this issue.

    """

    def __init__(self, error):
        self.error = error

    def __str__(self):
        return self.error


class DestinationExistsError(Error, RuntimeError):
    """The destination for a move or copy operation already exists.

    Parameters
    ----------
    destination :
        The destination object causing the error.

    """

    def __init__(self, destination):
        self.destination = destination


class JobsCorruptedError(Error, RuntimeError):
    """The state point manifest file of one or more jobs cannot be opened or is corrupted.

    Parameters
    ----------
    job_ids :
        The job id(s) of the corrupted job(s).

    """

    def __init__(self, job_ids):
        self.job_ids = job_ids


class StatepointParsingError(Error, RuntimeError):
    """Indicates an error that occurred while trying to identify a state point."""

    pass


class IncompatibleSchemaVersion(Error):
    """The project's schema version is incompatible with this version of signac."""

    pass



# Copyright (c) 2017 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""Synchronization of jobs and projects.

Jobs may be synchronized by copying all data from the source job to the
destination job. This means all files are copied and the documents
are synchronized. Conflicts, that means both jobs contain conflicting
data, may be resolved with a user defined strategy.

The synchronization of projects is in essence the synchronization of all jobs
which are in the destination project with the ones in the source project and
the sync synchronization of the project document. If a specific job does not
exist yet at the destination it is simply cloned, otherwise it is synchronized.

A sync strategy is a function (or functor) that takes the source job,
the destination job, and the name of the file generating the conflict
as arguments and returns the decision whether to overwrite the file as
Boolean. There are some default strategies defined within this module as
part of the :class:`~.FileSync` class. These are the default strategies:

    1. always -- Always overwrite on conflict.
    2. never -- Never overwrite on conflict.
    3. update -- Overwrite when the modification time of the source file is newer.
    4. Ask -- Ask the user interactively about each conflicting filename.

For example, to synchronize two projects resolving conflicts by modification time, use:

.. code-block:: python

    dest_project.sync(source_project, strategy=sync.FileSync.update)

Unlike files, which are always either overwritten as a whole or not, documents
can be synchronized more fine-grained with a *sync function*. Such a function (or
functor) takes the source and the destination document as arguments and performs
the synchronization. The user is encouraged to implement their own sync functions,
but there are a few default functions implemented as part of the :class:`~.DocSync` class:

    1. NO_SYNC -- Do not perform any synchronization.
    2. COPY -- Apply the same strategy used to resolve file conflicts.
    3. update -- Equivalent to dst.update(src).
    4. ByKey -- Synchronize the source document key by key, more information below.

This is how we could synchronize two jobs, where the documents are synchronized
with a simple update function:

.. code-block:: python

    dst_job.sync(src_job, doc_sync=sync.DocSync.update)

The :class:`.DocSync.ByKey` functor attempts to synchronize the destination document
with the source document without overwriting any data. That means this function
behaves similar to :func:`~.DocSync.update` for a non-intersecting set of keys,
but in addition will preserve nested mappings without overwriting values. In addition,
any key conflict, that means keys that are present in both documents, but have
differing data, will lead to the raise of a :class:`.DocumentSyncConflict` exception.
The user may expclitly decide to overwrite certain keys by providing a "key-strategy",
which is a function that takes the conflicting key as argument, and returns the
decision whether to overwrite that specific key as Boolean. For example, to sync
two jobs, where conflicting keys should only be overwritten if they contain the
term 'foo', we could execute:

.. code-block:: python

    dst_job.sync(src_job, doc_sync=sync.DocSync.ByKey(lambda key: 'foo' in key))

This means that all documents are synchronized 'key-by-key' and only conflicting keys that
contain the word "foo" will be overwritten, any other conflicts would lead to the
raise of a :class:`~.DocumentSyncConflict` exception. A key-strategy may also be
a regular expression, so the synchronization above could also be achieved with:

.. code-block:: python

    dst_job.sync(src_job, doc_sync=sync.DocSync.ByKey('foo'))
"""
import os
import re
from collections import defaultdict as ddict
from collections import namedtuple
from collections.abc import Mapping
from multiprocessing.pool import ThreadPool

from .contrib.utility import query_yes_no
from .errors import (
    DestinationExistsError,
    DocumentSyncConflict,
    FileSyncConflict,
    SchemaSyncConflict,
)
from .syncutil import _FileModifyProxy, dircmp, dircmp_deep, logger

__all__ = [
    "FileSync",
    "DocSync",
    "sync_jobs",
    "sync_projects",
]


# Definition of default sync strategies


class FileSync:
    """Collection of file synchronization strategies."""

    @classmethod
    def keys(cls):
        """Return keys."""
        return ("always", "never", "update", "Ask")

    @staticmethod
    def always(src, dst, fn):
        """Resolve sync conflicts by always overwriting."""
        return True

    @staticmethod
    def never(src, dst, fn):
        """Resolve sync conflicts by never overwriting."""
        return False

    @staticmethod
    def update(src, dst, fn):
        """Resolve sync conflicts based on newest modified timestamp."""
        return os.path.getmtime(src.fn(fn)) > os.path.getmtime(dst.fn(fn))

    class Ask:
        """Resolve sync conflicts by asking whether a file should be overwritten interactively."""

        def __init__(self):
            self.yes = set()
            self.no = set()

        def __call__(self, src, dst, fn):
            """Ask user if a file should be overwritten."""
            if fn in self.yes:
                return True
            elif fn in self.no:
                return False
            else:
                overwrite = query_yes_no(f"Overwrite files named '{fn}'?", "no")
                if overwrite:
                    self.yes.add(fn)
                    return True
                else:
                    self.no.add(fn)
                    return False


class DocSync:
    """Collection of document synchronization functions."""

    NO_SYNC = False
    "Do not synchronize documents."

    COPY = "copy"
    "Copy (and potentially overwrite) documents like any other file."

    @staticmethod
    def update(src, dst):
        """Perform a simple update."""
        for key in src.keys():
            dst[key] = src[key]

    class ByKey:
        """Synchronize documents key by key."""

        def __init__(self, key_strategy=None):
            if isinstance(key_strategy, str):

                def regex_key_strategy(key):
                    """Match keys according to key_strategy."""
                    return re.match(key_strategy, key)

                self.key_strategy = regex_key_strategy
            else:
                self.key_strategy = key_strategy
            self.skipped_keys = set()

        def __str__(self):
            return "{}({})".format(type(self).__name__, self.key_strategy)

        def __call__(self, src, dst, root=""):
            """Synchronize src and dst."""
            if src == dst:
                return
            for key, value in src.items():
                if key in dst:
                    if dst[key] == value:
                        continue
                    elif isinstance(value, Mapping):
                        self(src[key], dst[key], key + ".")
                        continue
                    elif self.key_strategy is None or not self.key_strategy(root + key):
                        self.skipped_keys.add(root + key)
                        continue
                dst[key] = value

            # Check for skipped keys and raise an exception in case that no strategy
            # was provided, otherwise just log them.
            if self.skipped_keys and not root:
                if self.key_strategy is None:
                    raise DocumentSyncConflict(self.skipped_keys)
                else:
                    logger.more("Skipped keys: {}".format(", ".join(self.skipped_keys)))


def _sync_job_workspaces(
    src, dst, strategy, exclude, copy, copytree, recursive=True, deep=False, subdir=""
):
    """Synchronize two job workspaces file by file, following the provided strategy."""
    if deep:
        diff = dircmp_deep(src.fn(subdir), dst.fn(subdir))
    else:
        diff = dircmp(src.fn(subdir), dst.fn(subdir))

    for fn in diff.left_only:
        if exclude and any([re.match(p, fn) for p in exclude]):
            logger.debug(f"File named '{fn}' is skipped (excluded).")
            continue
        fn_src = os.path.join(src.workspace(), subdir, fn)
        fn_dst = os.path.join(dst.workspace(), subdir, fn)
        if os.path.isfile(fn_src):
            copy(fn_src, fn_dst)
        elif recursive:
            copytree(fn_src, fn_dst)
        else:
            logger.warning(f"Skip directory '{fn_src}'.")
    for fn in diff.diff_files:
        if exclude and any([re.match(p, fn) for p in exclude]):
            logger.debug(f"File named '{fn}' is skipped (excluded).")
            continue
        if strategy is None:
            raise FileSyncConflict(fn)
        else:
            fn_src = os.path.join(src.workspace(), subdir, fn)
            fn_dst = os.path.join(dst.workspace(), subdir, fn)
            if strategy(src, dst, os.path.join(subdir, fn)):
                copy(fn_src, fn_dst)
            else:
                logger.debug(f"Skip file '{fn}'.")
    for _subdir in diff.subdirs:
        if recursive:
            _sync_job_workspaces(
                src=src,
                dst=dst,
                strategy=strategy,
                exclude=exclude,
                copy=copy,
                copytree=copytree,
                recursive=recursive,
                deep=deep,
                subdir=os.path.join(subdir, _subdir),
            )
        else:
            logger.warning("Skip directory '{}'.".format(os.path.join(subdir, _subdir)))


def _identical_path(a, b):
    """Verify if two absolute real paths match."""
    return os.path.abspath(os.path.realpath(a)) == os.path.abspath(os.path.realpath(b))


def sync_jobs(
    src,
    dst,
    strategy=None,
    exclude=None,
    doc_sync=None,
    recursive=False,
    follow_symlinks=True,
    preserve_permissions=False,
    preserve_times=False,
    preserve_owner=False,
    preserve_group=False,
    deep=False,
    dry_run=False,
):
    """Synchronize the dst job with the src job.

    By default, this method will synchronize all files and document data
    of dst job with the src job until a synchronization conflict occurs.
    There are two different kinds of synchronization conflicts:

        1. The two jobs have files with the same name, but different content.
        2. The two jobs have documents that share keys, but those keys are
           mapped to different values.

    A file conflict can be resolved by providing a 'FileSync' *strategy* or by
    *excluding* files from the synchronization. An unresolvable conflict is indicated
    with the raise of a :class:`~.errors.FileSyncConflict` exception.

    A document synchronization conflict can be resolved by providing a doc_sync function
    that takes the source and the destination document as first and second argument.

    Parameters
    ----------
    src : :class:`~signac.contrib.job.Job`
        The src job, data will be copied from this job's workspace.
    dst : :class:`~signac.contrib.job.Job`
        The dst job, data will be copied to this job's workspace.
    strategy : callable
        A synchronization strategy for file conflicts. The strategy should be a
        callable with signature ``strategy(src, dst, filepath)`` where ``src``
        and ``dst`` are the source and destination instances of
        :py:class:`~signac.Project` and ``filepath`` is the filepath relative
        to the project root. If no strategy is provided, a
        :class:`.errors.SyncConflict` exception will be raised upon conflict.
        (Default value = None)
    exclude : str
        A filename exclusion pattern. All files matching this pattern will be
        excluded from the synchronization process. (Default value = None)
    doc_sync : attribute or callable from :py:class:`~signac.sync.DocSync`
        A synchronization strategy for document keys. The default is to use a
        safe key-by-key strategy that will not overwrite any values on
        conflict, but instead raises a :class:`~.errors.DocumentSyncConflict`
        exception.
    recursive : bool
        Recursively synchronize sub-directories encountered within the job
        workspace directories. (Default value = False)
    follow_symlinks : bool
        Follow and copy the target of symbolic links. (Default value = True)
    preserve_permissions : bool
        Preserve file permissions (Default value = False)
    preserve_times : bool
        Preserve file modification times (Default value = False)
    preserve_owner : bool
        Preserve file owner (Default value = False)
    preserve_group : bool
        Preserve file group ownership (Default value = False)
    dry_run : bool
        If True, do not actually perform any synchronization operations.
        (Default value = False)
    deep : bool
        (Default value = False)

    """
    # Check identity
    if _identical_path(src.workspace(), dst.workspace()):
        raise ValueError("Source and destination can't be the same!")

    # check src and dst compatiblity
    assert src.FN_MANIFEST == dst.FN_MANIFEST
    assert src.FN_DOCUMENT == dst.FN_DOCUMENT

    # Nothing to be done if the src is not initialized.
    if src not in src._project:
        return

    # The doc_sync functions defaults to a safe "by_key" strategy.
    if doc_sync is None:
        doc_sync = DocSync.ByKey()

    # the exclude argument must be a list
    if exclude is None:
        exclude = []
    elif not isinstance(exclude, list):
        exclude = [exclude]
    exclude.append(src.FN_MANIFEST)
    if doc_sync != DocSync.COPY:
        exclude.append(src.FN_DOCUMENT)

    if type(dry_run) == _FileModifyProxy:
        proxy = dry_run
    else:
        proxy = _FileModifyProxy(
            root=src.workspace(),
            follow_symlinks=follow_symlinks,
            permissions=preserve_permissions,
            times=preserve_times,
            owner=preserve_owner,
            group=preserve_group,
            dry_run=bool(dry_run),
        )
    if proxy.dry_run:
        logger.debug(f"Synchronizing job '{src}' (dry run)...")
    else:
        logger.debug(f"Synchronizing job '{src}'...")

    if os.path.isdir(src.workspace()):
        if not dry_run:
            dst.init()
        _sync_job_workspaces(
            src=src,
            dst=dst,
            strategy=strategy,
            exclude=exclude,
            copy=proxy.copy,
            copytree=proxy.copytree,
            recursive=recursive,
            deep=deep,
        )

    if not (doc_sync is DocSync.NO_SYNC or doc_sync == DocSync.COPY):
        if src.document != dst.document:
            with proxy.create_doc_backup(dst.document) as dst_proxy:
                doc_sync(src.document, dst_proxy)


FileTransferStats = namedtuple("_FileTransferStats", ["num_files", "volume"])


def sync_projects(
    source,
    destination,
    strategy=None,
    exclude=None,
    doc_sync=None,
    selection=None,
    check_schema=True,
    recursive=False,
    follow_symlinks=True,
    preserve_permissions=False,
    preserve_times=False,
    preserve_owner=False,
    preserve_group=False,
    deep=False,
    dry_run=False,
    parallel=False,
    collect_stats=False,
):
    """Synchronize the destination project with the source project.

    Try to clone all jobs from the source to the destination.
    If the destination job already exist, try to synchronize the job using the
    optionally specified strategy.

    Parameters
    ----------
    source : class:`~.Project`
        The project presenting the source for synchronization.
    destination : class:`~.Project`
        The project that is modified for synchronization.
    strategy : callable
        A synchronization strategy for file conflicts. The strategy should be a
        callable with signature ``strategy(src, dst, filepath)`` where ``src``
        and ``dst`` are the source and destination instances of
        :py:class:`~signac.Project` and ``filepath`` is the filepath relative
        to the project root. If no strategy is provided, a
        :class:`.errors.SyncConflict` exception will be raised upon conflict.
        (Default value = None)
    exclude : str
        A filename exclusion pattern. All files matching this pattern will be
        excluded from the synchronization process. (Default value = None)
    doc_sync : attribute or callable from :py:class:`~signac.sync.DocSync`
        A synchronization strategy for document keys. The default is to use a
        safe key-by-key strategy that will not overwrite any values on
        conflict, but instead raises a :class:`~.errors.DocumentSyncConflict`
        exception.
    selection : sequence of :class:`~signac.contrib.job.Job` or job ids (str)
        Only synchronize the given selection of jobs. (Default value = None)
    check_schema : bool
        If True, only synchronize if this and the other project have a matching
        state point schema. See also: :meth:`~.detect_schema`. (Default value =
        True)
    recursive : bool
        Recursively synchronize sub-directories encountered within the job
        workspace directories. (Default value = False)
    follow_symlinks : bool
        Follow and copy the target of symbolic links. (Default value = True)
    preserve_permissions : bool
        Preserve file permissions (Default value = False)
    preserve_times : bool
        Preserve file modification times (Default value = False)
    preserve_owner : bool
        Preserve file owner (Default value = False)
    preserve_group : bool
        Preserve file group ownership (Default value = False)
    dry_run : bool
        If True, do not actually perform the synchronization operation, just
        log what would happen theoretically. Useful to test synchronization
        strategies without the risk of data loss. (Default value = False)
    deep : bool
        (Default value = False)
    parallel : bool
        (Default value = False)
    collect_stats : bool
        (Default value = False)

    Returns
    -------
    NoneType or :class:`~signac.sync.FileTransferStats`
        Returns stats if ``collect_stats`` is ``True``, else ``None``.

    Raises
    ------
    DocumentSyncConflict
        If there are conflicting keys within the project or job documents that
        cannot be resolved with the given strategy or if there is no strategy
        provided.
    FileSyncConflict
        If there are differing files that cannot be resolved with the given
        strategy or if no strategy is provided.
    SchemaSyncConflict
        In case that the check_schema argument is True and the detected state
        point schema of this and the other project differ.

    """
    if source == destination:
        raise ValueError("Source and destination project cannot be identical!")

    # Setup data modification proxy
    proxy = _FileModifyProxy(
        root=source.workspace(),
        follow_symlinks=follow_symlinks,
        permissions=preserve_permissions,
        times=preserve_times,
        owner=preserve_owner,
        group=preserve_group,
        dry_run=dry_run,
        collect_stats=collect_stats,
    )

    # Perform a schema check in an attempt to avoid bad sync operations.
    if check_schema:
        schema_src = source.detect_schema()
        schema_dst = destination.detect_schema()
        if schema_dst and schema_src and schema_src != schema_dst:
            only_in_src = schema_src.difference(schema_dst)
            only_in_dst = schema_dst.difference(schema_src)
            if only_in_src or only_in_dst:
                raise SchemaSyncConflict(schema_src, schema_dst)

    if doc_sync is None:
        doc_sync = DocSync.ByKey()

    if (
        selection is not None
    ):  # The selection argument may be a jobs or job ids sequence.
        selection = {str(j) for j in selection}

    # Provide some information about this sync process.
    if selection:
        logger.info(
            "Synchronizing selection ({}) of project '{}' to '{}'.".format(
                len(selection), source, destination
            )
        )
    else:
        logger.info(f"Synchronizing project '{source}' to '{destination}'.")
    logger.more(f"'{source.root_directory()}' -> '{destination.root_directory()}'")
    if dry_run:
        logger.info("Performing dry run!")
    if exclude is not None:
        logger.more(f"File name exclude pattern: '{exclude}'")
    logger.more(f"Sync strategy: '{strategy}'")
    logger.more(f"Doc sync strategy: '{doc_sync}'")

    # Sync the Project document.
    if not (doc_sync is DocSync.NO_SYNC or doc_sync == DocSync.COPY):
        if source.document != destination.document:
            with proxy.create_doc_backup(destination.document) as dst_proxy:
                doc_sync(source.document, dst_proxy)

    # Sync jobs from source to destination.
    logger.more("Collect all jobs to synchronize...")
    if selection is None:
        jobs_to_sync = list(source)
    else:
        jobs_to_sync = [job for job in source if job.get_id() in selection]

    N = len(jobs_to_sync)
    logger.more(f"Synchronizing {N} jobs.")
    count = ddict(int)

    def _clone_or_sync(src_job):
        """Clone a job if it does not exist, or sync if it exists."""
        try:
            destination.clone(src_job, copytree=proxy.copytree)
            logger.more(f"Cloned job '{src_job}'.")
            return 1
        except DestinationExistsError:
            dst_job = destination.open_job(id=src_job.get_id())
            sync_jobs(
                src=src_job,
                dst=dst_job,
                strategy=strategy,
                exclude=exclude,
                doc_sync=doc_sync,
                recursive=recursive,
                dry_run=proxy,  # used as internal argument to forward the proxy
            )
            logger.more(f"Synchonized job '{src_job}'.")
            return 2

    if parallel:
        num_processes = None if parallel is True else parallel
        logger.more(
            "Parallelizing over {} threads for synchronization.".format(
                "multiple" if num_processes is None else num_processes
            )
        )
        with ThreadPool(None if parallel is True else parallel) as pool:
            for i, ret in enumerate(pool.imap(_clone_or_sync, jobs_to_sync)):
                count[ret] += 1
                logger.info("Project sync progress: {}/{}".format(i + 1, N))
    else:
        for i, src_job in enumerate(jobs_to_sync):
            count[_clone_or_sync(src_job)] += 1
            logger.info("Project sync progress: {}/{}".format(i + 1, N))

    num_cloned, num_synchronized = count[1], count[2]
    logger.info(f"Cloned {num_cloned} and synchronized {num_synchronized} job(s).")
    if collect_stats:
        return FileTransferStats(**proxy.stats)



# Copyright (c) 2019 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""Define signac version."""


__version__ = "1.5.1"

SCHEMA_VERSION = "1"


__all__ = [
    "__version__",
    "SCHEMA_VERSION",
]



# Copyright (c) 2017 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
import errno
import hashlib
import importlib.machinery
import logging
import math
import os
import re
import warnings
from collections import defaultdict
from time import sleep

from deprecation import deprecated

from ..common import errors
from ..core import json
from ..version import __version__
from .hashing import calc_id
from .utility import walkdepth

logger = logging.getLogger(__name__)

KEY_PROJECT = "project"
KEY_FILENAME = "filename"
KEY_PATH = "root"
KEY_PAYLOAD = "format"

"""
THIS MODULE IS DEPRECATED!
"""


@deprecated(
    deprecated_in="1.3",
    removed_in="2.0",
    current_version=__version__,
    details="The indexing module is deprecated.",
)
def md5(file):
    "Calculate and return the md5 hash value for the file data."
    m = hashlib.md5()
    for chunk in iter(lambda: file.read(4096), b""):
        m.update(chunk)
    return m.hexdigest()


def _is_blank_module(module):
    with open(module.__file__) as file:
        return not bool(file.read().strip())


# this class is deprecated
class BaseCrawler:
    """Crawl through `root` and index all files.

    The crawler creates an index on data, which can be exported
    to a database for easier access."""

    tags = None

    @deprecated(
        deprecated_in="1.3",
        removed_in="2.0",
        current_version=__version__,
        details="The indexing module is deprecated.",
    )
    def __init__(self, root):
        """Initialize a BaseCrawler instance.

        :param root: The path to the root directory to crawl through.
        :type root: str"""
        self.root = os.path.expanduser(root)
        self.tags = set() if self.tags is None else set(self.tags)

    def docs_from_file(self, dirpath, fn):
        """Implement this method to generate documents from files.

        :param dirpath: The path of the file, relative to `root`.
        :type dirpath: str
        :param fn: The filename.
        :type fn: str
        :yields: Index documents.
        """
        raise NotImplementedError()
        yield

    def fetch(self, doc, mode="r"):
        """Implement this generator method to associate data with a document.

        :returns: object associated with doc
        """
        raise errors.FetchError(f"Unable to fetch object for '{doc}'.")

    @classmethod
    def _calculate_hash(cls, doc, dirpath, fn):
        blob = json.dumps(doc, sort_keys=True)
        m = hashlib.md5()
        m.update(dirpath.encode("utf-8"))
        m.update(fn.encode("utf-8"))
        m.update(blob.encode("utf-8"))
        return m.hexdigest()

    def crawl(self, depth=0):
        """Crawl through the `root` directory.

        The crawler will inspect every file and directory up
        until the specified `depth` and call the
        :meth:`docs_from_file` method.

        :param depth: Crawl through the directory for the specified depth.
                      A value of 0 specifies no limit.
        :type dept: int
        :yields: (id, doc)-tuples"""
        logger.info(f"Crawling '{self.root}' (depth={depth})...")
        for dirpath, dirnames, filenames in walkdepth(self.root, depth):
            for fn in filenames:
                for doc in self.docs_from_file(dirpath, fn):
                    logger.debug(
                        "doc from file: '{}'.".format(os.path.join(dirpath, fn))
                    )
                    doc.setdefault(KEY_PAYLOAD, None)
                    doc.setdefault("_id", self._calculate_hash(doc, dirpath, fn))
                    yield doc
        logger.info(f"Crawl of '{self.root}' done.")

    def process(self, doc, dirpath, fn):
        """Implement this method for additional processing of generated docs.

        The default implementation will return the unmodified `doc`.

        :param dirpath: The path of the file, relative to `root`.
        :type dirpath: str
        :param fn: The filename.
        :type fn: str
        :returns: A document, that means an instance of mapping.
        :rtype: mapping"""
        return doc


# this class is deprecated
class RegexFileCrawler(BaseCrawler):
    r"""Generate documents from filenames and associate each file with a data type.

    The `RegexFileCrawler` uses regular expressions to generate
    data from files. This is a particular easy method to retrieve meta data
    associated with files. Inherit from this class to configure a crawler
    for your data structre.

    Let's assume we want to index text files, with a naming pattern, that
    specifies a parameter `a` through the filename, e.g.:

    .. code-block:: python

        ~/my_project/a_0.txt
        ~/my_project/a_1.txt
        ...

    A valid regular expression to match this pattern would
    be: ``.*\/a_(?P<a>\d+)\.txt`` which may be defined for a crawler as such:

    .. code-block:: python

        MyCrawler(RegexFileCrawler):
            pass

        MyCrawler.define('.*\/a_(?P<a>\d+)\.txt', 'TextFile')
    """
    "Mapping of compiled regex objects and associated formats."
    definitions = {}  # type: ignore

    @classmethod
    @deprecated(
        deprecated_in="1.3",
        removed_in="2.0",
        current_version=__version__,
        details="The indexing module is deprecated.",
    )
    def define(cls, regex, format_=None):
        """Define a format for a particular regular expression.

        :param regex: All files of the specified format
            must match this regular expression.
        :type regex: :class:`str`
        :param format_: The format associated with all matching files.
        :type format_: :class:`object`
        """
        if isinstance(regex, str):
            regex = re.compile(regex)
        definitions = dict(cls.definitions)
        definitions[regex] = format_
        cls.definitions = definitions

    @classmethod
    @deprecated(
        deprecated_in="1.3",
        removed_in="2.0",
        current_version=__version__,
        details="The indexing module is deprecated.",
    )
    def compute_file_id(cls, doc, file):
        """Compute the file id for a given doc and the associated file.

        :param doc: The index document
        :param file: The associated file
        :returns: The file id.
        """
        file_id = doc["md5"] = md5(file)
        return file_id

    @deprecated(
        deprecated_in="1.3",
        removed_in="2.0",
        current_version=__version__,
        details="The indexing module is deprecated.",
    )
    def docs_from_file(self, dirpath, fn):
        """Generate documents from filenames.

        This method implements the abstract
        :py:meth:~.BaseCrawler.docs_from_file` and yields index
        documents associated with files.

        .. note::
            It is not recommended to reimplement this method to modify
            documents generated from filenames.
            See :meth:`~RegexFileCrawler.process` instead.

        :param dirpath: The path of the file relative to root.
        :param fn: The filename of the file.
        :yields: Index documents.
        """
        for regex, format_ in self.definitions.items():
            m = regex.match(os.path.join(dirpath, fn))
            if m:
                doc = self.process(m.groupdict(), dirpath, fn)
                doc[KEY_FILENAME] = os.path.relpath(
                    os.path.join(dirpath, fn), self.root
                )
                doc[KEY_PATH] = os.path.abspath(self.root)
                doc[KEY_PAYLOAD] = str(format_)
                with open(os.path.join(dirpath, fn), "rb") as file:
                    doc["file_id"] = self.compute_file_id(doc, file)
                yield doc

    @deprecated(
        deprecated_in="1.3",
        removed_in="2.0",
        current_version=__version__,
        details="The indexing module is deprecated.",
    )
    def fetch(self, doc, mode="r"):
        """Fetch the data associated with `doc`.

        :param doc: A index document.
        :type doc: :class:`dict`
        :returns: The file associated with the index document.
        :rtype: A file-like object
        """
        fn = doc.get(KEY_FILENAME)
        if fn:
            for regex, format_ in self.definitions.items():
                ffn = os.path.join(self.root, fn)
                m = regex.match(ffn)
                if m:
                    if isinstance(format_, str):
                        return open(ffn, mode=mode)
                    else:
                        for meth in ("read", "close"):
                            if not callable(getattr(format_, meth, None)):
                                msg = f"Format {format_} has no {meth}() method."
                                warnings.warn(msg)
                        return format_(open(ffn, mode=mode))
            else:
                raise errors.FetchError(
                    f"Unable to match file path of doc '{doc}' to format definition."
                )
        else:
            raise errors.FetchError(f"Insufficient meta data in doc '{doc}'.")

    @deprecated(
        deprecated_in="1.3",
        removed_in="2.0",
        current_version=__version__,
        details="The indexing module is deprecated.",
    )
    def process(self, doc, dirpath, fn):
        """Post-process documents generated from filenames.

        Example:

        .. code-block:: python

            MyCrawler(signac.indexing.RegexFileCrawler):
                def process(self, doc, dirpath, fn):
                    doc['long_name_for_a'] = doc['a']
                    return super(MyCrawler, self).process(doc, dirpath, fn)

        :param dirpath: The path of the file, relative to `root`.
        :type dirpath: str
        :param fn: The filename.
        :type fn: str
        :returns: An index document, that means an instance of mapping.
        :rtype: mapping
        """
        result = {}
        for key, value in doc.items():
            if value is None or isinstance(value, bool):
                result[key] = value
                continue
            try:
                value = float(value)
            except Exception:
                result[key] = value
            else:
                if not math.isnan(value) or math.isinf(value):
                    if float(value) == int(value):
                        result[key] = int(value)
                    else:
                        result[key] = float(value)
        return super().process(result, dirpath, fn)

    @deprecated(
        deprecated_in="1.3",
        removed_in="2.0",
        current_version=__version__,
        details="The indexing module is deprecated.",
    )
    def crawl(self, depth=0):
        if self.definitions:
            yield from super().crawl(depth=depth)
        else:
            return


@deprecated(
    deprecated_in="1.3",
    removed_in="2.0",
    current_version=__version__,
    details="The indexing module is deprecated.",
)
class JSONCrawler(BaseCrawler):
    encoding = "utf-8"
    fn_regex = r".*\.json"

    def docs_from_json(self, doc):
        yield doc

    def docs_from_file(self, dirpath, fn):
        if re.match(self.fn_regex, os.path.join(dirpath, fn)):
            with open(os.path.join(dirpath, fn), "rb") as file:
                doc = json.loads(file.read().decode(self.encoding))
                yield from self.docs_from_json(doc)


def _index_signac_project_workspace(
    root,
    include_job_document=True,
    fn_statepoint="signac_statepoint.json",
    fn_job_document="signac_job_document.json",
    statepoint_index="statepoint",
    signac_id_alias="_id",
    encoding="utf-8",
    statepoint_dict=None,
):
    "Yields standard index documents for a signac project workspace."
    logger.debug(f"Indexing workspace '{root}'...")
    m = re.compile(r"[a-f0-9]{32}")
    try:
        job_ids = [jid for jid in os.listdir(root) if m.match(jid)]
    except OSError as error:
        if error.errno == errno.ENOENT:
            return
        else:
            raise
    for i, job_id in enumerate(job_ids):
        if not m.match(job_id):
            continue
        doc = {"signac_id": job_id, KEY_PATH: root}
        if signac_id_alias:
            doc[signac_id_alias] = job_id
        fn_sp = os.path.join(root, job_id, fn_statepoint)
        with open(fn_sp, "rb") as file:
            sp = json.loads(file.read().decode(encoding))
            if statepoint_dict is not None:
                statepoint_dict[job_id] = sp
            if statepoint_index:
                doc[statepoint_index] = sp
            else:
                doc.update(sp)
        if include_job_document:
            fn_doc = os.path.join(root, job_id, fn_job_document)
            try:
                with open(fn_doc, "rb") as file:
                    doc.update(json.loads(file.read().decode(encoding)))
            except OSError as error:
                if error.errno != errno.ENOENT:
                    raise
        yield doc
    if job_ids:
        logger.debug("Indexed workspace '{}', {} entries.".format(root, i + 1))


# this class is deprecated
class SignacProjectCrawler(RegexFileCrawler):
    """Index a signac project workspace.

    Without any file format definitions, this crawler
    yields index documents for each job, including
    the statepoint and the job document.

    See also: :py:class:`~.RegexFileCrawler`

    :param root: The path to the project's root directory.
    :type root: str"""

    encoding = "utf-8"
    statepoint_index = "statepoint"
    fn_statepoint = "signac_statepoint.json"
    fn_job_document = "signac_job_document.json"
    signac_id_alias = "_id"

    @deprecated(
        deprecated_in="1.3",
        removed_in="2.0",
        current_version=__version__,
        details="The indexing module is deprecated.",
    )
    def __init__(self, root):
        from .project import get_project

        root = get_project(root=root).workspace()
        self._statepoints = {}
        return super().__init__(root=root)

    def _get_job_id(self, dirpath):
        return os.path.relpath(dirpath, self.root).split("/")[0]

    def _read_statepoint(self, job_id):
        fn_sp = os.path.join(self.root, job_id, self.fn_statepoint)
        with open(fn_sp, "rb") as file:
            return json.loads(file.read().decode(self.encoding))

    def _get_statepoint(self, job_id):
        sp = self._statepoints.setdefault(job_id, self._read_statepoint(job_id))
        assert calc_id(sp) == job_id
        return sp

    def get_statepoint(self, dirpath):
        job_id = self._get_job_id(dirpath)
        return job_id, self._get_statepoint(self, job_id)

    def process(self, doc, dirpath, fn):
        if dirpath is not None:
            job_id = self._get_job_id(dirpath)
            statepoint = self._get_statepoint(job_id)
            doc["signac_id"] = job_id
            if self.statepoint_index:
                doc[self.statepoint_index] = statepoint
            else:
                doc.update(statepoint)
        return super().process(doc, dirpath, fn)

    def crawl(self, depth=0):
        for doc in _index_signac_project_workspace(
            root=self.root,
            fn_statepoint=self.fn_statepoint,
            fn_job_document=self.fn_job_document,
            statepoint_index=self.statepoint_index,
            signac_id_alias=self.signac_id_alias,
            encoding=self.encoding,
            statepoint_dict=self._statepoints,
        ):
            yield self.process(doc, None, None)
        for doc in super().crawl(depth=depth):
            yield doc


# this class is deprecated
class MainCrawler(BaseCrawler):
    r"""Compiles a main index from indexes defined in access modules.

    An instance of this crawler will search the data space for access
    modules, which by default are named ``signac_access.py``. Once such
    a file is found, the crawler will import the module and try to execute
    two special functions given that they are defined within the module's
    global namespace: ``get_indexes()`` and ``get_crawlers()``.

    The ``get_indexes()`` is assumed to yield one or multiple index generator
    functions, while the ``get_crawlers()`` function is assumed to yield
    one or more crawler instances.

    This is an example for such an access module:

    .. code-block:: python

        import signac

        def get_indexes(root):
            yield signac.index_files(root, r'.*\.txt')

        def get_crawlers(root):
            yield MyCrawler(root)

    In case that the main crawler has tags, the ``get_indexes()`` function
    will always be ignored while crawlers yielded from the ``get_crawlers()``
    function will only be executed in case that they match at least one
    of the tags.

    In case that the access module is completely empty, it will be executed
    as if it had the following directives:

    .. code-block:: python

        import signac

        def get_indexes(root):
            yield signac.get_project(root).index()

    Tags for indexes yielded from the `get_indexes()` function can be specified
    by assigning them directly to the function:

    .. code-block:: python

        def get_indexes(root):
            yield signac.get_project(root).index()

        get_indexes.tags = {'foo'}


    :param root: The path to the root directory to crawl through.
    :type root: str
    :param raise_on_error: Raise all exceptions encountered during
        during crawling instead of ignoring them.
    :type raise_on_error: bool
    """

    FN_ACCESS_MODULE = "signac_access.py"
    "The filename of modules containing crawler definitions."

    @deprecated(
        deprecated_in="1.3",
        removed_in="2.0",
        current_version=__version__,
        details="The indexing module is deprecated.",
    )
    def __init__(self, root, raise_on_error=False):
        self.raise_on_error = raise_on_error
        super().__init__(root=root)

    def _docs_from_module(self, dirpath, fn):
        name = os.path.join(dirpath, fn)
        module = importlib.machinery.SourceFileLoader(name, name).load_module()

        logger.info(f"Crawling from module '{module.__file__}'.")

        has_tags = self.tags is not None and len(set(self.tags))

        def _check_tags(tags):
            if tags is None or not len(set(tags)):
                if has_tags:
                    logger.info("Skipping, index has no defined tags.")
                    return False
                else:
                    return True
            else:
                if not has_tags:
                    logger.info("Skipping, index requires tags.")
                    return False
                elif set(self.tags).intersection(set(tags)):
                    return True  # at least one tag matches!
                else:
                    logger.info("Skipping, tag mismatch.")
                    return False

        if not has_tags and _is_blank_module(module):
            from .project import get_project

            for doc in get_project(root=dirpath).index():
                yield doc

        if hasattr(module, "get_indexes"):
            if _check_tags(getattr(module.get_indexes, "tags", None)):
                for index in module.get_indexes(dirpath):
                    for doc in index:
                        yield doc

        if hasattr(module, "get_crawlers"):
            for crawler in module.get_crawlers(dirpath):
                logger.info(f"Executing subcrawler:\n {crawler}")
                if _check_tags(getattr(crawler, "tags", None)):
                    for doc in crawler.crawl():
                        doc.setdefault(KEY_PROJECT, os.path.relpath(dirpath, self.root))
                        yield doc

    def docs_from_file(self, dirpath, fn):
        """Compile main index from file in case it is an access module.

        :param dirpath: The path of the file relative to root.
        :param fn: The filename of the file.
        :yields: Index documents.
        """
        if fn == self.FN_ACCESS_MODULE:
            try:
                yield from self._docs_from_module(dirpath, fn)
            except Exception:
                logger.error(
                    "Error while indexing from module '{}'.".format(
                        os.path.join(dirpath, fn)
                    )
                )
                if self.raise_on_error:
                    raise
            else:
                logger.debug(
                    "Completed indexing from '{}'.".format(os.path.join(dirpath, fn))
                )


# Deprecated API
class MasterCrawler(MainCrawler):
    def __init__(self, *args, **kwargs):
        warnings.warn(
            "The MasterCrawler class has been replaced by the MainCrawler class. "
            "Both classes are deprecated and will be removed in a future release.",
            DeprecationWarning,
        )
        super().__init__(*args, **kwargs)


@deprecated(
    deprecated_in="1.3",
    removed_in="2.0",
    current_version=__version__,
    details="The indexing module is deprecated.",
)
def fetch(
    doc_or_id, mode="r", mirrors=None, num_tries=3, timeout=60, ignore_local=False
):
    """Fetch the file associated with this document or file id.

    This function retrieves a file associated with the provided
    index document or file id and behaves like the built-in
    :py:func:`open` function, e.g.:

    .. code-block:: python

        for doc in index:
            with signac.fetch(doc) as file:
                do_something_with(file)

    :param doc_or_id: A file_id or a document with a file_id value.
    :param mode: Mode to use for opening files.
    :param mirrors: An optional set of mirrors to fetch the file from.
    :param num_tries: The number of automatic retry attempts in case of
        mirror connection errors.
    :type num_tries: int
    :param timeout: The time in seconds to wait before an
        automatic retry attempt.
    :type timeout: int
    :returns: The file associated with the document or file id.
    :rtype: A file-like object
    """
    if doc_or_id is None:
        raise ValueError("Argument 'doc_or_id' must not be None!")
    file_id = doc_or_id if isinstance(doc_or_id, str) else doc_or_id.get("file_id")
    if not ignore_local:
        try:
            fn = os.path.join(doc_or_id["root"], doc_or_id["filename"])
            return open(fn, mode=mode)
        except KeyError:
            raise errors.FetchError("Insufficient file meta data for fetch.", doc_or_id)
        except OSError as error:
            if error.errno == errno.ENOENT:
                if file_id is None:
                    raise errors.FetchError(f"Failed to fetch '{doc_or_id}'.")
    if mirrors is None:
        raise errors.FetchError("No mirrors provided!")
    else:
        for i in range(num_tries):
            for mirror in mirrors:
                try:
                    return mirror.get(file_id, mode=mode)
                except mirror.AutoRetry as error:
                    logger.warning(error)
                    sleep(timeout)
                except mirror.FileNotFoundError as error:
                    logger.debug(error)
            else:
                raise errors.FetchError(f"Unable to fetch object for '{file_id}'.")


@deprecated(
    deprecated_in="1.3",
    removed_in="2.0",
    current_version=__version__,
    details="The indexing module is deprecated.",
)
def fetched(docs):
    """Iterate over documents and yield associated files."""
    for doc in docs:
        if "file_id" in doc:
            yield doc, fetch(doc)


def _export_to_mirror(file, file_id, mirror):
    "Export a file-like object with file_id to mirror."
    with mirror.new_file(_id=file_id) as dst:
        dst.write(file.read())


@deprecated(
    deprecated_in="1.3",
    removed_in="2.0",
    current_version=__version__,
    details="The indexing module is deprecated.",
)
def export_to_mirror(doc, mirror, num_tries=3, timeout=60):
    """Export a file associated with doc to mirror.

    :param doc: A document with a file_id entry.
    :param mirror: A file-system object to export the file to.
    :param num_tries: The number of automatic retry attempts in case of
        mirror connection errors.
    :type num_tries: int
    :param timeout: The time in seconds to wait before an
        automatic retry attempt.
    :type timeout: int
    :returns: The file id after successful export.
    """
    if "file_id" not in doc:
        raise errors.ExportError(f"Doc '{doc}' does not have a file_id entry.")
    for i in range(num_tries):
        try:
            with fetch(doc, mode="rb") as file:
                _export_to_mirror(file, doc["file_id"], mirror)
        except mirror.FileExistsError:
            logger.debug(
                "File with id '{}' already exported, skipping.".format(doc["file_id"])
            )
            break
        except mirror.AutoRetry as error:
            logger.warning(f"Error during export: '{error}', retrying...")
            sleep(timeout)
        else:
            logger.debug(
                "Stored file with id '{}' in mirror '{}'.".format(
                    doc["file_id"], mirror
                )
            )
            return doc["file_id"]
    else:
        raise errors.ExportError(doc)


@deprecated(
    deprecated_in="1.3",
    removed_in="2.0",
    current_version=__version__,
    details="The indexing module is deprecated.",
)
def export_one(doc, index, mirrors=None, num_tries=3, timeout=60):
    """Export one document to index and an optionally associated file to mirrors.

    :param doc: A document with a file_id entry.
    :param docs: The index collection to export to.
    :param mirrors: An optional set of mirrors to export files to.
    :param num_tries: The number of automatic retry attempts in case of
        mirror connection errors.
    :type num_tries: int
    :param timeout: The time in seconds to wait before an
        automatic retry attempt.
    :type timeout: int
    :returns: The id and file id after successful export.
    """
    index.replace_one({"_id": doc["_id"]}, doc, upsert=True)
    if mirrors and "file_id" in doc:
        for mirror in mirrors:
            export_to_mirror(doc, mirror, num_tries, timeout)
        return doc["_id"], doc["file_id"]
    else:
        return doc["_id"], None


@deprecated(
    deprecated_in="1.3",
    removed_in="2.0",
    current_version=__version__,
    details="The indexing module is deprecated.",
)
def export(docs, index, mirrors=None, update=False, num_tries=3, timeout=60, **kwargs):
    """Export docs to index and optionally associated files to mirrors.

    The behavior of this function is equivalent to:

    .. code-block:: python

        for doc in docs:
            export_one(doc, index, mirrors, num_tries)

    If the `update` argument is set to True, the export algorithm will
    automatically identify stale index documents, that means documents
    that refer to files or state points that have been removed and are
    no longer part of the data space. Any document which shares the
    `root`, but not the `_id` field with any of the updated documents
    is considered stale and removed. Using `update` in combination with
    an empty docs sequence will raise `ExportError`, since it is not
    possible to identify stale documents in that case.

    .. note::

        This function will automatically delegate to specialized
        implementations for special index types. For example, if
        the index argument is a MongoDB document collection, the
        index documents will be exported via :py:func:`~.export_pymongo`.

    :param docs: The index documents to export.
    :param index: The collection to export the index to.
    :param mirrors: An optional set of mirrors to export files to.
    :param update: If True, remove stale index documents, that means
        documents that refer to files or state points that no longer exist.
    :type update: bool
    :param num_tries: The number of automatic retry attempts in case of
        mirror connection errors.
    :type num_tries: int
    :param timeout: The time in seconds to wait before an
        automatic retry attempt.
    :type timeout: int
    :param kwargs: Optional keyword arguments to pass to
        delegate implementations.
    :raises ExportError: When using the update argument in combination with
        an empty docs sequence.
    """
    try:
        import pymongo
    except ImportError:
        pass
    else:
        if isinstance(index, pymongo.collection.Collection):
            logger.info("Using optimized export function export_pymongo().")
            return export_pymongo(
                docs=docs,
                index=index,
                mirrors=mirrors,
                update=update,
                num_tries=num_tries,
                timeout=timeout,
                **kwargs,
            )
    ids = defaultdict(list)
    for doc in docs:
        _id, _ = export_one(doc, index, mirrors, num_tries, timeout, **kwargs)
        if update:
            root = doc.get("root")
            if root is not None:
                ids[root].append(_id)
    if update:
        if ids:
            stale = set()
            for root in ids:
                docs_ = index.find({"root": root})
                all_ = {doc["_id"] for doc in docs_}
                stale.update(all_.difference(ids[root]))
            logger.info("Removing {} stale documents.".format(len(stale)))
            for _id in set(stale):
                index.delete_one(dict(_id=_id))
        else:
            raise errors.ExportError(
                "The exported docs sequence is empty! Unable to update!"
            )


def _export_pymongo(docs, operations, index, mirrors, num_tries, timeout):
    """Export docs via operations to index and files to mirrors."""
    import pymongo

    if mirrors is not None:
        for mirror in mirrors:
            for doc in docs:
                if "file_id" in doc:
                    export_to_mirror(doc, mirror, num_tries, timeout)
    for i in range(num_tries):
        try:
            index.bulk_write(operations)
            break
        except pymongo.errors.AutoReconnect as error:
            logger.warning(error)
            sleep(timeout)
    else:
        raise errors.ExportError()


@deprecated(
    deprecated_in="1.3",
    removed_in="2.0",
    current_version=__version__,
    details="The indexing module is deprecated.",
)
def export_pymongo(
    docs, index, mirrors=None, update=False, num_tries=3, timeout=60, chunksize=100
):
    """Optimized :py:func:`~.export` function for pymongo index collections.

    The behavior of this function is rougly equivalent to:

    .. code-block:: python

        for doc in docs:
            export_one(doc, index, mirrors, num_tries)

    .. note::

        All index documents must be JSON-serializable to
        be able to be exported to a MongoDB collection.

    :param docs: The index documents to export.
    :param index: The database collection to export the index to.
    :type index: :class:`pymongo.collection.Collection`
    :param num_tries: The number of automatic retry attempts in case of
        mirror connection errors.
    :type num_tries: int
    :param timeout: The time in seconds to wait before an
        automatic retry attempt.
    :type timeout: int
    :param chunksize: The buffer size for export operations.
    :type chunksize: int"""
    import pymongo

    logger.info(f"Exporting to pymongo database collection index '{index}'.")
    chunk = []
    operations = []
    ids = defaultdict(list)
    for doc in docs:
        f = {"_id": doc["_id"]}
        if update:
            root = doc.get("root")
            if root is not None:
                ids[root].append(doc["_id"])
        chunk.append(doc)
        operations.append(pymongo.ReplaceOne(f, doc, upsert=True))
        if len(chunk) >= chunksize:
            logger.debug("Pushing chunk.")
            _export_pymongo(chunk, operations, index, mirrors, num_tries, timeout)
            chunk[:] = []
            operations[:] = []
    if len(operations):
        logger.debug("Pushing final chunk.")
        _export_pymongo(chunk, operations, index, mirrors, num_tries, timeout)
    if update:
        if ids:
            stale = set()
            for root in ids:
                docs_ = index.find({"root": root})
                all_ = {doc["_id"] for doc in docs_}
                stale.update(all_.difference(ids[root]))
            logger.info("Removing {} stale documents.".format(len(stale)))
            for _id in set(stale):
                index.delete_one(dict(_id=_id))
        else:
            raise errors.ExportError(
                "The exported docs sequence is empty! Unable to update!"
            )


@deprecated(
    deprecated_in="1.3",
    removed_in="2.0",
    current_version=__version__,
    details="The indexing module is deprecated.",
)
def index_files(root=".", formats=None, depth=0):
    r"""Generate a file index.

    This generator function yields file index documents,
    where each index document corresponds to one file.

    To index all files in the current working directory,
    simply execute:

    .. code-block:: python

        for doc in signac.index_files():
            print(doc)

    A file associated with a file index document can be
    fetched via the :py:func:`fetch` function:

    .. code-block:: python

        for doc in signac.index_files():
            with signac.fetch(doc) as file:
                print(file.read())

    This is especially useful if the file index is part of
    a collection (:py:class:`.Collection`) which can be searched
    for specific entries.

    To limit the file index to files with a specific filename
    formats, provide a regular expression as the formats argument.
    To index all files that have file ending `.txt`, execute:

    .. code-block:: python

        for doc in signac.index_files(formats='.*\.txt'):
            print(doc)

    We can specify specific formats by providing a dictionary as
    ``formats`` argument, where the key is the filename pattern and
    the value is an arbitrary formats string, e.g.:

    .. code-block:: python

        for doc in signac.index_files(formats=
            {r'.*\.txt': 'TextFile', r'.*\.zip': 'ZipFile'}):
            print(doc)

    :param root: The directory to index, defaults to the
        current working directory.
    :type root: str
    :param formats: Limit the index to files that match the
        given regular expression and optionally associate formats
        with given patterns.
    :param depth: Limit the search to the specified directory depth.
    :type depth: int
    :yields: The file index documents as dicts.
    """
    if formats is None:
        formats = {".*": "File"}
    if isinstance(formats, str):
        formats = {formats: "File"}

    class Crawler(RegexFileCrawler):
        pass

    for regex, fmt in formats.items():
        Crawler.define(regex, fmt)

    yield from Crawler(root).crawl(depth=depth)


@deprecated(
    deprecated_in="1.3",
    removed_in="2.0",
    current_version=__version__,
    details="The indexing module is deprecated.",
)
def index(root=".", tags=None, depth=0, **kwargs):
    r"""Generate a main index.

    A main index is compiled from other indexes by searching
    for modules named ``signac_access.py`` and compiling all
    indexes which are yielded from a function ``get_indexes(root)``
    defined within that module as well as the indexes generated by
    crawlers yielded from a function ``get_crawlers(root)`` defined
    within that module.

    This is a minimal example for a ``signac_access.py`` file:

    .. code-block:: python

        import signac

        def get_indexes(root):
            yield signac.index_files(root, r'.*\.txt')

    Internally, this function constructs an instance of
    :py:class:`.MainCrawler` and all extra key-word arguments
    will be forwarded to the constructor of said main crawler.

    :param root: Look for access modules under this directory path.
    :type root: str
    :param tags: If tags are provided, do not execute subcrawlers
        that don't match the same tags.
    :param depth: Limit the search to the specified directory depth.
    :param kwargs: These keyword-arguments are forwarded to the
        internal MainCrawler instance.
    :type depth: int
    :yields: The main index documents as instances of dict.
    """

    class Crawler(MainCrawler):
        pass

    if tags is not None:
        Crawler.tags = tags

    yield from Crawler(root, **kwargs).crawl(depth=depth)



# Copyright (c) 2019 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""Basic wrapper to access multiple different data stores."""

import errno
import os
import re
import uuid


class DictManager:
    """Helper class to manage multiple instances of dict-like classes.

    This class is designed to manage multiple dict-like interface classes to files
    with a shared prefix (directory).
    """

    cls = None
    suffix = None

    __slots__ = ["_prefix", "_dict_registry"]

    def __init__(self, prefix):
        assert (
            self.cls is not None
        ), "Subclasses of DictManager must define the cls variable."
        assert (
            self.suffix is not None
        ), "Subclasses of DictManager must define the suffix variable."
        self._prefix = os.path.abspath(prefix)
        self._dict_registry = {}

    @property
    def prefix(self):
        """Return the prefix."""
        return self._prefix

    def __eq__(self, other):
        return (
            os.path.realpath(self.prefix) == os.path.realpath(other.prefix)
            and self.suffix == other.suffix
        )

    def __repr__(self):
        return "{}(prefix={})".format(
            type(self).__name__, repr(os.path.relpath(self.prefix))
        )

    __str__ = __repr__

    def __getitem__(self, key):
        if key not in self._dict_registry:
            self._dict_registry[key] = self.cls(
                os.path.join(self.prefix, key) + self.suffix
            )
        return self._dict_registry[key]

    @staticmethod
    def _validate_key(key):
        """Emit a warning or raise an exception if key is invalid. Returns key."""
        return key

    def __setitem__(self, key, value):
        self._validate_key(key)
        tmp_key = str(uuid.uuid4())
        try:
            self[tmp_key].update(value)
            os.replace(self[tmp_key].filename, self[key].filename)
        except OSError as error:
            if error.errno == errno.ENOENT and not len(value):
                raise ValueError("Cannot assign empty value!")
            else:
                raise error
        except Exception as error:
            try:
                del self[tmp_key]
            except KeyError:
                pass
            raise error
        else:
            del self._dict_registry[key]

    def __delitem__(self, key):
        try:
            os.unlink(self[key].filename)
        except OSError as error:
            if error.errno == errno.ENOENT:
                raise KeyError(key)
            else:
                raise error

    def __getattr__(self, name):
        try:
            return super().__getattribute__(name)
        except AttributeError:
            if name.startswith("__") or name in self.__slots__:
                raise
            try:
                return self.__getitem__(name)
            except KeyError:
                raise AttributeError(name)

    def __setattr__(self, name, value):
        if name.startswith("__") or name in self.__slots__:
            super().__setattr__(name, value)
        else:
            self.__setitem__(name, value)

    def __delattr__(self, name):
        if name.startswith("__") or name in self.__slots__:
            super().__delattr__(name)
        else:
            self.__delitem__(name)

    def __iter__(self):
        for fn in os.listdir(self.prefix):
            m = re.match(f"^(.*){self.suffix}$", fn)
            if m:
                yield m.groups()[0]

    def keys(self):
        """Return an iterable of keys."""
        return iter(self)

    def __len__(self):
        return len(list(self.keys()))

    def __getstate__(self):
        return dict(_prefix=self._prefix, _dict_registry=self._dict_registry)

    def __setstate__(self, d):
        self._prefix = d["_prefix"]
        self._dict_registry = d["_dict_registry"]



# The MIT License (MIT)
#
# Copyright (c) 2014, 2015 Adrian Price-Whelan & Dan Foreman-Mackey
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""MPIPool for MPI-based multiprocessing-like process pools.

This 3rd party module is copied from https://github.com/adrn/mpipool."""

from deprecation import deprecated

__all__ = ["MPIPool"]
__version__ = "0.0.1"

"""
THIS MODULE IS DEPRECATED!
"""


@deprecated(
    deprecated_in="1.3",
    removed_in="2.0",
    current_version=__version__,
    details="The mpipool module is deprecated.",
)
class MPIPool:
    """
    A pool that distributes tasks over a set of MPI processes using
    mpi4py. MPI is an API for distributed memory parallelism, used
    by large cluster computers. This class provides a similar interface
    to Python's multiprocessing Pool, but currently only supports the
    :func:`map` method.

    Contributed initially by `Joe Zuntz <https://github.com/joezuntz>`_.

    Parameters

    comm : (optional)
        The ``mpi4py`` communicator.

    debug : bool (optional)
        If ``True``, print out a lot of status updates at each step.

    loadbalance : bool (optional)
        if ``True`` and the number of tasks is greater than the
        number of processes, tries to loadbalance by sending out
        one task to each cpu first and then sending out the rest
        as the cpus get done.
    """

    def __init__(self, comm=None, debug=False, loadbalance=False):
        if comm is None:
            # Late import of the MPI constant is necessary, to avoid
            # early mpi initialization, which causes critital errors
            # on badly configured systems.
            from mpi4py import MPI

            self.comm = MPI.COMM_WORLD
        else:
            self.comm = comm
        self.rank = self.comm.Get_rank()
        self.size = self.comm.Get_size() - 1
        self.debug = debug
        self.function = _error_function
        self.loadbalance = loadbalance
        if self.size == 0:
            raise ValueError(
                "Tried to create an MPI pool, but there "
                "was only one MPI process available. "
                "Need at least two."
            )

    def is_master(self):
        """
        Is the current process the master?

        """
        return self.rank == 0

    def wait(self):
        """
        If this isn't the master process, wait for instructions.

        """
        from mpi4py import MPI

        if self.is_master():
            raise RuntimeError("Master node told to await jobs.")

        status = MPI.Status()

        while True:
            # Event loop.
            # Sit here and await instructions.
            if self.debug:
                print(f"Worker {self.rank} waiting for task.")

            # Blocking receive to wait for instructions.
            task = self.comm.recv(source=0, tag=MPI.ANY_TAG, status=status)
            if self.debug:
                print(f"Worker {self.rank} got task {task} with tag {status.tag}.")

            # Check if message is special sentinel signaling end.
            # If so, stop.
            if isinstance(task, _close_pool_message):
                if self.debug:
                    print(f"Worker {self.rank} told to quit.")
                break

            # Check if message is special type containing new function
            # to be applied
            if isinstance(task, _function_wrapper):
                self.function = task.function
                if self.debug:
                    print(
                        f"Worker {self.rank} replaced its task function: {self.function}."
                    )
                continue

            # If not a special message, just run the known function on
            # the input and return it asynchronously.
            result = self.function(task)
            if self.debug:
                print(
                    f"Worker {self.rank} sending answer {result} with tag {status.tag}."
                )
            self.comm.isend(result, dest=0, tag=status.tag)

    def map(self, function, tasks, ntask=None, callback=None):
        """
        Like the built-in :func:`map` function, apply a function to all
        of the values in a list and return the list of results.

        Parameters

        function : callable
            The function to apply to each element in the list.

        tasks :
            A list of tasks -- each element is passed to the input
            function.

        callback : callable (optional)
            A callback function to call on each result.

        """
        from mpi4py import MPI

        if ntask is None:
            ntask = len(tasks)

        # If not the master just wait for instructions.
        if not self.is_master():
            self.wait()
            return

        if function is not self.function:
            if self.debug:
                print(f"Master replacing pool function with {function}.")

            self.function = function
            F = _function_wrapper(function)

            # Tell all the workers what function to use.
            requests = []
            for i in range(self.size):
                r = self.comm.isend(F, dest=i + 1)
                requests.append(r)

            # Wait until all of the workers have responded. See:
            #       https://gist.github.com/4176241
            MPI.Request.waitall(requests)

        if (not self.loadbalance) or (ntask <= self.size):
            # Do not perform load-balancing - the default load-balancing
            # scheme emcee uses.

            # Send all the tasks off and wait for them to be received.
            # Again, see the bug in the above gist.
            requests = []
            for i, task in enumerate(tasks):
                worker = i % self.size + 1
                if self.debug:
                    print(f"Sent task {task} to worker {worker} with tag {i}.")
                r = self.comm.isend(task, dest=worker, tag=i)
                requests.append(r)

            MPI.Request.waitall(requests)

            # Now wait for the answers.
            results = []
            for i in range(ntask):
                worker = i % self.size + 1
                if self.debug:
                    print(f"Master waiting for worker {worker} with tag {i}")
                result = self.comm.recv(source=worker, tag=i)

                if callback is not None:
                    callback(result)

                results.append(result)

            return results

        else:
            # Perform load-balancing. The order of the results are likely to
            # be different from the previous case.
            for i, task in enumerate(tasks[0 : self.size]):
                worker = i + 1
                if self.debug:
                    print(f"Sent task {task} to worker {worker} with tag {i}.")
                # Send out the tasks asynchronously.
                self.comm.isend(task, dest=worker, tag=i)

            ntasks_dispatched = self.size
            results = [None] * ntask
            for itask in range(ntask):
                status = MPI.Status()
                # Receive input from workers.
                result = self.comm.recv(
                    source=MPI.ANY_SOURCE, tag=MPI.ANY_TAG, status=status
                )
                worker = status.source
                i = status.tag

                if callback is not None:
                    callback(result)

                results[i] = result
                if self.debug:
                    print(f"Master received from worker {worker} with tag {i}")

                # Now send the next task to this idle worker (if there are any
                # left).
                if ntasks_dispatched < ntask:
                    task = tasks[ntasks_dispatched]
                    i = ntasks_dispatched
                    if self.debug:
                        print(f"Sent task {task} to worker {worker} with tag {i}.")
                    # Send out the tasks asynchronously.
                    self.comm.isend(task, dest=worker, tag=i)
                    ntasks_dispatched += 1

            return results

    def bcast(self, *args, **kwargs):
        """
        Equivalent to mpi4py :func:`bcast` collective operation.
        """
        return self.comm.bcast(*args, **kwargs)

    def close(self):
        """
        Just send a message off to all the pool members which contains
        the special :class:`_close_pool_message` sentinel.

        """
        if self.is_master():
            for i in range(self.size):
                self.comm.isend(_close_pool_message(), dest=i + 1)

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()


class _close_pool_message:
    def __repr__(self):
        return "<Close pool message>"


class _function_wrapper:
    def __init__(self, function):
        self.function = function


def _error_function(task):
    raise RuntimeError("Pool was sent tasks before being told what function to apply.")



# Copyright (c) 2018 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.

# The API of the Collection class is adapted from the API of
# the pymongo.Collection class provided as part of the
# mongo-python-driver library [1], which is licensed under the
# Apache License 2.0.
#
# The implementation found here is designed to replicate a subset
# of the behavior of a pymongo.Collection while operating directly
# on files on the local file system instead of a MongoDB database.
#
# [1]: https://github.com/mongodb/mongo-python-driver
"""Collection in signac defined here."""

import argparse
import io
import logging
import operator
import re
import sys
from itertools import islice
from math import isclose
from numbers import Number

from ..core import json
from .filterparse import parse_filter_arg
from .utility import _nested_dicts_to_dotted_keys, _to_hashable

logger = logging.getLogger(__name__)


_INDEX_OPERATORS = (
    "$eq",
    "$gt",
    "$gte",
    "$lt",
    "$lte",
    "$ne",
    "$in",
    "$nin",
    "$regex",
    "$type",
    "$where",
    "$near",
)

_TYPES = {
    "int": int,
    "float": float,
    "bool": bool,
    "str": str,
    "list": tuple,
    "null": type(None),
}


MAX_DEFAULT_ID = int("F" * 32, 16)


def _flatten(container):
    """Yield elements from the container.

    Parameters
    ----------
    container : sequence
        List or tuple of elements.

    Yields
    ------
    Elements of container, recursively flattening sub-sequences.

    """
    for i in container:
        if isinstance(i, (list, tuple)):
            yield from _flatten(i)
        else:
            yield i


class _DictPlaceholder:
    pass


def _valid_filter(f, top=True):
    """Return True if the filter is valid.

    Parameters
    ----------
    f : dict
        The filter argument.
    top : bool
        Used to return True if the filter passed is of type dict.
        (Default value = True)

    Returns
    -------
    bool
        True if filter is valid.

    """
    if f is None:
        return True
    elif type(f) is dict:
        return all(_valid_filter(v, top=False) for v in f.values())
    elif type(f) is list:
        return not top
    else:
        return True


class _float(float):
    # Numerical objects of either integer or float type, that share the same numerical value,
    # but not the same type, are distinguished within a Collection, but considered equal within
    # Python. We manipulate the hash value, to enable the storage of both an int and a float
    # that share the same numerical value within a collection index (dict).

    # There is no risk of accidentally equating ints and floats with different values, since the
    # hash equality is only a necessary, not a sufficient condition for equality.
    def __hash__(self):
        return super().__hash__() + 1


class _TypedSetDefaultDict(dict):
    """Dictionary that is guaranteed to store differently typed values separately.

    This is necessary, because the hash value of integers with float type is identical
    to the same integer as int type, which means they cannot be stored separately in a
    standard dict.

    """

    def keys(self):
        for key in dict.keys(self):
            yield float(key) if type(key) is _float else key

    __iter__ = keys

    def items(self):
        for key, value in dict.items(self):
            yield float(key) if type(key) is _float else key, value

    def __missing__(self, key):
        value = set()
        dict.__setitem__(self, key, value)
        return value

    def __getitem__(self, key):
        return dict.__getitem__(self, _float(key) if type(key) is float else key)

    def __setitem__(self, key, value):
        return dict.__setitem__(self, _float(key) if type(key) is float else key, value)

    def __delitem__(self, key):
        dict.__delitem__(self, _float(key) if type(key) is float else key)

    def get(self, key, default=None):
        """Get the value for given key.

        Parameters
        ----------
        key : str, float
            The key to get the value.
        default :
            Default value if type of key is not float (Default value = None).

        Returns
        -------
        The value for given key.

        """
        return dict.get(self, _float(key) if type(key) is float else key, default)


def _build_index(docs, key, primary_key):
    """Build an index for 'key'; highly performance critical code path.

    Parameters
    ----------
    docs : iterable
        iterable of doc to build index.
    key : str
        The key to build index.
    primary_key : str
        The primary key.

    Returns
    -------
    :class:`~_TypedSetDefaultDict`
        Index for key.

    Raises
    ------
    InvalidKeyError
        The document contains invalid keys.

    """
    nodes = key.split(".")
    index = _TypedSetDefaultDict()

    for doc in docs:
        try:
            v = doc[nodes[0]]
            for n in nodes[1:]:
                v = v[n]
            if type(v) is dict:
                v = _DictPlaceholder
        except (KeyError, TypeError):
            pass
        except Exception as error:
            raise RuntimeError(
                f"An unexpected error occurred while processing doc '{doc}': {error}."
            )
        else:
            # inlined for performance
            if type(v) is dict:
                continue
            elif type(v) is list:  # performance
                index[_to_hashable(v)].add(doc[primary_key])
            else:
                index[v].add(doc[primary_key])

        if len(nodes) > 1:
            try:
                v = doc[".".join(nodes)]
            except KeyError:
                pass
            else:
                from ..errors import InvalidKeyError

                raise InvalidKeyError(
                    "\nThe document contains invalid keys. "
                    "Specifically keys with dots ('.').\n\n"
                    "See https://signac.io/document-wide-migration/ "
                    "for a recipe on how to replace dots in existing keys."
                )
    return index


def _find_with_index_operator(index, op, argument):
    """Find index for given operator and argument.

    Parameters
    ----------
    index : dict
        Index for the operator.
    op : str
        logical operator.
    argument :
        Dependent on the choice of logical operator argument (op).
        For better understanding have a look at :meth:`~Collection.find`.

    Returns
    -------
    set
        Index for given operator and argument.

    Raises
    ------
    ValueError
        When unknown argument is given for $type operator (When the operator is $type).

    """
    if op == "$in":

        def op(value, argument):
            return value in argument

    elif op == "$nin":

        def op(value, argument):
            return value not in argument

    elif op == "$regex":

        def op(value, argument):
            if isinstance(value, str):
                return re.search(argument, value)
            else:
                return False

    elif op == "$type":

        def op(value, argument):
            if argument in _TYPES:
                t = _TYPES[argument]
            else:
                raise ValueError(f"Unknown argument for $type operator: '{argument}'.")
            return isinstance(value, t)

    elif op == "$where":

        def op(value, argument):
            return eval(argument)(value)

    elif op == "$near":
        rel_tol, abs_tol = 1e-9, 0.0  # default values
        if isinstance(argument, (list, tuple)):
            if len(argument) == 1:
                argument = argument[0]
            elif len(argument) == 2:
                argument, rel_tol = argument
            elif len(argument) == 3:
                argument, rel_tol, abs_tol = argument
            else:
                err_msg = "The argument of the $near operator must be a float "
                err_msg += "or a list of floats with length 1, 2, or 3."
                raise ValueError(err_msg)
        argument = float(argument)
        rel_tol = float(rel_tol)
        abs_tol = float(abs_tol)

        def op(value, argument):
            return isclose(value, argument, rel_tol=rel_tol, abs_tol=abs_tol)

    else:
        op = getattr(operator, {"$gte": "$ge", "$lte": "$le"}.get(op, op)[1:])
    matches = set()
    for value in index:
        if op(value, argument):
            matches.update(index[value])
    return matches


def _check_logical_operator_argument(op, argument):
    """Check arguments for the logical-operator.

    Parameters
    ----------
    op : str
        logical-operator.
    argument : list
        list of arguments for logical-operator.

    Raises
    ------
    ValueError
        The argument of logical-operator is not a list or is an empty list.

    """
    if not isinstance(argument, list):
        raise ValueError(f"The argument of logical-operator '{op}' must be a list!")
    if not len(argument):
        raise ValueError(f"The argument of logical-operator '{op}' cannot be empty!")


class _CollectionSearchResults:
    """Iterator for a Collection result vector."""

    def __init__(self, collection, _ids):
        self._collection = collection
        self._ids = _ids

    def __iter__(self):
        return (self._collection[_id] for _id in self._ids)

    def __len__(self):
        return len(self._ids)

    count = __len__


class JSONParseError(ValueError):
    """Error class for JSON Parse."""

    pass


class Collection:
    """A collection of documents.

    The Collection class manages a collection of documents in memory
    or in a file on disk. A document is defined as a dictionary mapping
    of key-value pairs.

    An instance of collection may be used to manage and search documents.
    For example, given a collection with member data, where each document
    contains a `name` entry and an `age` entry, we can find the name of
    all members that are at age 32 like this:

    .. code-block:: python

        members = [
            {'name': 'John',  'age': 32},
            {'name': 'Alice', 'age': 28},
            {'name': 'Kevin', 'age': 32},
            # ...
            ]

        member_collection = Collection(members)
        for doc in member_collection.find({'age': 32}):
            print(doc['name'])

    To iterate over all documents in the collection, use:

    .. code-block:: python

        for doc in collection:
            print(doc)

    By default a collection object will reside in memory. However, it is
    possible to manage a collection associated to a file on disk. To open
    a collection which is associated with a file on disk, use the
    :meth:`Collection.open` class method:

    .. code-block:: python

        with Collection.open('collection.txt') as collection:
            for doc in collection.find({'age': 32}):
                print(doc)

    The collection file is by default opened in `a+` mode, which means it can
    be read from and written to and will be created if it does not exist yet.

    Parameters
    ----------
    docs : iterable
        Initialize the collection with these documents.
    primary_key : str
        The name of the key which serves as the primary
        index of the collection. Selecting documents by primary key has
        time complexity of O(N) in the worst case and O(1) on average.
        All documents must have a primary key value. The default primary
        key is `_id`.
    compresslevel : int
        The level of compression to use. Any positive value
        implies compression and is used by the underlying gzip implementation.
        Default value is 0 (no compression).

    Raises
    ------
    ValueError
        When first argument is a string.

    """

    def __init__(self, docs=None, primary_key="_id", compresslevel=0, _trust=False):
        if isinstance(docs, str):
            raise ValueError(
                "First argument cannot be of str type. "
                "Did you mean to use {}.open()?".format(type(self).__name__)
            )
        self.index_rebuild_threshold = 0.1
        self._primary_key = primary_key
        if compresslevel > 0:
            self._file = io.BytesIO()
        else:
            self._file = io.StringIO()
        self._compresslevel = compresslevel
        self._requires_flush = False
        self._dirty = set()
        self._indexes = {}
        self._next_default_id_ = None
        self._docs = {}
        if docs is not None:
            for doc in docs:
                if self._primary_key not in doc:
                    doc[self._primary_key] = self._next_default_id()
                self.__setitem__(doc[self._primary_key], doc, _trust=_trust)
            self._update_indexes()

    def _assert_open(self):
        if self._docs is None:
            raise RuntimeError(
                "Trying to access closed {}.".format(type(self).__name__)
            )

    def _next_default_id(self):
        """Return next default id.

        Returns
        -------
        _id : str
            Next default id.

        Raises
        ------
        RuntimeError
            When unable to determine default id.

        """
        if self._next_default_id_ is None:
            self._next_default_id_ = len(self)
        for i in range(len(self) + 1):
            assert self._next_default_id_ < MAX_DEFAULT_ID
            _id = str(hex(self._next_default_id_))[2:].rjust(32, "0")
            self._next_default_id_ += 1
            if _id not in self:
                return _id
        raise RuntimeError("Unable to determine default id.")

    def _remove_from_indexes(self, _id):
        """Remove index corresponding to given id.

        Parameters
        ----------
        _id : str
            id to remove from the index.

        """
        for index in self._indexes.values():
            remove_keys = set()
            for key, group in index.items():
                if _id in group:  # faster than exception handling (performance)
                    group.remove(_id)
                if not len(group):
                    remove_keys.add(key)
            for key in remove_keys:
                del index[key]

    def _update_indexes(self):
        """Update the indexes."""
        if self._dirty:
            for _id in self._dirty:
                self._remove_from_indexes(_id)
            docs = [self[_id] for _id in self._dirty]
            for key, index in self._indexes.items():
                tmp = _build_index(docs, key, self._primary_key)
                for v, group in tmp.items():
                    index[v].update(group)
            self._dirty.clear()

    def _build_index(self, key):
        """Build index for given key.

        Parameters
        ----------
        key : str
            The key to build index for.

        """
        logger.debug(f"Building index for key '{key}'...")
        self._indexes[key] = _build_index(self._docs.values(), key, self._primary_key)
        logger.debug(f"Built index for key '{key}'.")

    def index(self, key, build=False):
        """Get (and optionally build) the index for a given key.

        An index allows to access documents by a specific key with
        minimal time complexity, e.g.:

        .. code-block:: python

            age_index = member_collection.index('age')
            for _id in age_index[32]:
                print(member_collection[_id]['name'])

        This means we can access documents by the 'age' key in O(1) time on
        average in addition to the primary key. Using the :meth:`.find`
        method will automatically build all required indexes for the particular
        search.

        Once an index has been built, it will be internally managed by the
        class and updated with subsequent changes. An index returned by this
        method is always current with the latest state of the collection.

        Parameters
        ----------
        key : str
            The primary key of the requested index.
        build : bool
            If True, build a non-existing index if necessary,
            otherwise raise KeyError (Default value = False).

        Returns
        -------
        dict
            Index for the given key.

        Raises
        ------
        KeyError
            In case the build is False and the index has not been built yet or
            no index is present for the key.

        """
        if key == self._primary_key:
            raise KeyError("Can't access index for primary key via index() method.")
        elif key in self._indexes:
            if len(self._dirty) > self.index_rebuild_threshold * len(self):
                logger.debug("Indexes outdated, rebuilding...")
                self._indexes.clear()
                self._build_index(key)
                self._dirty.clear()
            else:
                self._update_indexes()
        else:
            if build:
                self._build_index(key)
            else:
                raise KeyError(f"No index for key '{key}'.")
        return self._indexes[key]

    def __str__(self):
        return "<{} file={}>".format(type(self).__name__, self._file)

    def __iter__(self):
        try:
            return iter(self._docs.values())
        except AttributeError:
            raise RuntimeError(
                "Trying to access closed {}.".format(type(self).__name__)
            )

    @property
    def ids(self):
        """Get an iterator over the primary key in the collection.

        Returns
        -------
        iterable
            iterator over the primary key in the collection.

        """
        self._assert_open()
        return iter(self._docs)

    @property
    def primary_key(self):
        """Get the name of the collection's primary key (default='_id')."""
        return self._primary_key

    def __len__(self):
        self._assert_open()
        return len(self._docs)

    def __contains__(self, _id):
        return _id in self._docs

    def __getitem__(self, _id):
        # The _assert_open() check is only performed after an
        # exception has been caught, which is slightly faster
        # than running the check every time.
        try:
            return self._docs[_id].copy()
        except TypeError:
            self._assert_open()
            raise

    @staticmethod
    def _validate_key(key):
        """Emit a warning or raise an exception if key is invalid. Returns key.

        Parameters
        ----------
        key : str
            Key to validate.

        Returns
        -------
        str
            If key is validated.

        Raises
        ------
        InvalidKeyError
            When key given is invalid.

        """
        if "." in key:
            from ..errors import InvalidKeyError

            raise InvalidKeyError("Keys may not contain dots ('.').")
        return key

    @classmethod
    def _validate_doc(cls, doc):
        """Emit a warning or raise an exception if the document is invalid. Returns doc.

        Parameters
        ----------
        doc : dict
            The document to validate.

        Returns
        -------
        doc : dict
            If doc is validated.

        """
        try:
            for key in doc.keys():
                cls._validate_doc(doc[cls._validate_key(key)])
        except AttributeError:
            return
        return doc

    def __setitem__(self, _id, doc, _trust=False):
        self._assert_open()
        if not isinstance(_id, str):
            raise TypeError("The primary key must be of type str!")
        doc.setdefault(self._primary_key, _id)
        if _id != doc[self._primary_key]:
            raise ValueError("Primary key mismatch!")
        if _trust:
            self._docs[_id] = doc
        else:
            try:
                doc_ = json.loads(json.dumps(doc))
            except TypeError as error:
                raise TypeError(
                    f"Serialization of document '{doc}' failed with error: {error}"
                )
            self._docs[_id] = self._validate_doc(doc_)
        self._dirty.add(_id)
        self._requires_flush = True

    def insert_one(self, doc):
        """Insert one document into the collection.

        If the document does not have a value for the
        collection's primary key yet, it will be assigned one.

        .. code-block:: python

            _id = collection.insert_one(doc)
            assert _id in collection

        .. note::

            The document will be directly updated in case that
            it has no primary key and must therefore be mutable!

        Parameters
        ----------
        doc : dict
            The document to be inserted.

        Returns
        -------
        str
            The _id of the inserted documented.

        """
        self._assert_open()
        if self._primary_key in doc:
            _id = doc[self._primary_key]
        else:
            _id = doc[self._primary_key] = self._next_default_id()
        if _id in self:
            raise KeyError("Primary key collision!")
        self[_id] = doc
        return _id

    def __delitem__(self, _id):
        self._assert_open()
        del self._docs[_id]
        self._remove_from_indexes(_id)
        try:
            self._dirty.remove(_id)
        except KeyError:
            pass
        self._requires_flush = True

    def clear(self):
        """Remove all documents from the collection."""
        self._docs.clear()
        self._indexes.clear()
        self._dirty.clear()
        self._requires_flush = True

    def update(self, docs):
        """Update the collection with these documents.

        Any existing documents with the same primary key
        will be replaced.

        Parameters
        ----------
        docs : iterable
            A sequence of documents to be upserted into the collection.

        """
        for doc in docs:
            if self._primary_key in doc:
                _id = doc[self._primary_key]
            else:
                _id = doc[self._primary_key] = self._next_default_id()
            self[_id] = doc

    def _find_expression(self, key, value):
        """Find document for key value pair.

        Parameters
        ----------
        key : str
            The key for expression-operator.
        value :
            The value for expression-operator.

        Returns
        -------
        set
            The document for key value pair.

        Raises
        ------
        KeyError
            When Bad operator expression/ Bad operator placement or
            the expression-operator is unknown.
        ValueError
            The value is not bool when operator for '$exists' operator.

        """
        logger.debug(f"Find documents for expression '{key}: {value}'.")
        if "$" in key:
            if key.count("$") > 1:
                raise KeyError(f"Bad operator expression '{key}'.")
            nodes = key.split(".")
            op = nodes[-1]
            if not op.startswith("$"):
                raise KeyError(f"Bad operator placement '{key}'.")
            key = ".".join(nodes[:-1])
            if op in _INDEX_OPERATORS:
                index = self.index(key, build=True)
                return _find_with_index_operator(index, op, value)
            elif op == "$exists":
                if not isinstance(value, bool):
                    raise ValueError(
                        "The value of the '$exists' operator must be boolean."
                    )
                index = self.index(key, build=True)
                match = {elem for elems in index.values() for elem in elems}
                return match if value else set(self.ids).difference(match)
            else:
                raise KeyError(f"Unknown expression-operator '{op}'.")
        else:
            index = self.index(key, build=True)
            # Check to see if 'value' is a floating point type but an
            # integer value (e.g., 4.0), and search for both the int and float
            # values. This allows the user to find statepoints that have
            # integer-valued keys that are stored as floating point types.
            # Note that this both cases: 1) user searches for an int and hopes
            # to find values that are stored as integer-valued floats and 2) user
            # searches for a integer-valued float and hopes to find ints.
            # This way, both `signac find x 4.0` and `signac find x 4` would
            # return jobs where `sp.x` is stored as either 4.0 or 4.
            if isinstance(value, Number) and float(value).is_integer():
                result_float = index.get(_float(value), set())
                result_int = index.get(int(value), set())
                return result_int.union(result_float)
            else:
                return index.get(value, set())

    def _find_result(self, expr):
        """Find ids for given expression.

        Parameters
        ----------
        expr : str
            The expression for which to get ids.

        Returns
        -------
        set
            Set of all the ids if the given expression is empty.

        """
        if not len(expr):
            return set(self.ids)  # Empty expression yields all ids...

        result_ids = None

        def reduce_results(match):
            """Reduce the results by intersection of matches.

            Parameters
            ----------
            match : set
                match for the given expression.

            """
            nonlocal result_ids
            if result_ids is None:  # First match
                result_ids = match
            else:  # Update previous match
                result_ids = result_ids.intersection(match)

        # Check if filter contains primary key, in which case we can
        # immediately reduce the result.
        _id = expr.pop(self._primary_key, None)
        if _id is not None and _id in self:
            reduce_results({_id})

        # Extract all logical-operator expressions for now.
        or_expressions = expr.pop("$or", None)
        and_expressions = expr.pop("$and", None)
        not_expression = expr.pop("$not", None)

        # Reduce the result based on the remaining non-logical expression:
        for key, value in _nested_dicts_to_dotted_keys(expr):
            reduce_results(self._find_expression(key, value))
            if not result_ids:  # No match, no need to continue...
                return set()

        # Reduce the result based on the logical-operator expressions:
        if not_expression is not None:
            not_match = self._find_result(not_expression)
            reduce_results(set(self.ids).difference(not_match))

        if and_expressions is not None:
            _check_logical_operator_argument("$and", and_expressions)
            for expr_ in and_expressions:
                reduce_results(self._find_result(expr_))

        if or_expressions is not None:
            _check_logical_operator_argument("$or", or_expressions)
            or_results = set()
            for expr_ in or_expressions:
                or_results.update(self._find_result(expr_))
            reduce_results(or_results)

        assert result_ids is not None
        return result_ids

    def _find(self, filter=None, limit=0):
        """Return a result vector of ids for the given filter and limit.

        This function normalizes the filter argument and then attempts to
        build a result vector for the given key-value queries.
        For each key that is queried, an internal index is built and then
        searched.

        The result vector is a set of ids, where each id is the value of the
        primary key of a document that matches the given filter.

        The _find() method uses the following optimizations:

            1. If the filter is None, the result is directly returned, since
               all documents will match an empty filter.
            2. If the filter argument contains a primary key, the result
               is directly returned since no search operation is necessary.
            3. The filter is processed key by key, once the result vector is
               empty it is immediately returned.

        Parameters
        ----------
        filter : dict
            The filter argument that all documents must match (Default value = None).
        limit : int
            Limit the size of the result vector (Default value = 0).

        Returns
        -------
        set
            A set of ids of documents that match the given filter.

        Raises
        ------
        ValueError
            When the filter argument is invalid.

        """
        self._assert_open()
        if filter:
            filter = json.loads(json.dumps(filter))  # Normalize
            if not _valid_filter(filter):
                raise ValueError(filter)
            result = self._find_result(filter)
            return set(islice(result, limit if limit else None))
        else:
            return set(islice(self._docs.keys(), limit if limit else None))

    def find(self, filter=None, limit=0):
        """Find all documents matching filter, but not more than limit.

        This function searches the collection for all documents that match
        the given filter and returns a result vector. For example:

        .. code-block:: python

            for doc in collection.find(my_filter):
                print(doc)

        Nested values should be searched using the ``.`` operator, for example:

        .. code-block:: python

            docs = collection.find({'nested.value': 42})

        will return documents with a nested structure: ``{'nested': {'value': 42}}``.

        The result of :meth:`~.find` can be stored and iterated over multiple times.
        In addition, the result vector can be queried for its size:

        .. code-block:: python

            docs = collection.find(my_filter)

            print(len(docs))    # the number of documents matching

            for doc in docs:    # iterate over the result vector
                pass

        Arithmetic Operators

                * *$eq*: equal
                * *$ne*: not equal
                * *$gt*: greater than
                * *$gte*: greater or equal than
                * *$lt*: less than
                * *$lte*: less or equal than

            .. code-block:: python

                project.find({"a": {"$lt": 5})

            Matches all docs with *a* less than 5.

        Logical Operators

            That includes *$and* and *$or*; both expect a list of expressions.

                .. code-block:: python

                    project.find({"$or": [{"a": 4}, {"b": {"$gt": 3}}]})

            Matches all docs, where *a* is 4 or *b* is greater than 3.

        Exists operator

            Determines whether a specific key exists, or not, e.g.:

                    .. code-block:: python

                        project.find({"a": {"$exists": True}})

        Array operator

            To determine whether specific elements are in (*$in*), or not in (*$nin*)
            an array, e.g.:

                    .. code-block:: python

                        project.find({"a": {"$in": [0, 1, 2]}})

            Matches all docs, where *a* is either 0, 1, or 2. Usage of *$nin* is equivalent.

        Regular expression operator

            Allows the "on-the-fly" evaluation of regular expressoions, e.g.:

                    .. code-block:: python

                        project.find({"protocol": {"$regex": "foo"}})

            Will match all docs with a protocol that contains the term 'foo'.

        $type operator

            Matches when a value is of specific type, e.g.:

                    .. code-block:: python

                        project.find({"protocol": {"$type": str}})

            Finds all docs, where the value of protocol is of type str.
            Other types that can be checked are: *int*, *float*, *bool*, *list*, and *null*.

        $where operator

            Matches an arbitrary python expression, e.g.:

                    .. code-block:: python

                        project.find({"foo": {"$where": "lambda x: x.startswith('bar')"}})

            Matches all docs, where the value for foo starts with the word 'bar'.

        Parameters
        ----------
        filter : dict
            All documents must match the given filter (Default value = None).
        limit : int
            Do not return more than limit number of documents.
            A limit value of 0 (the default) means no limit.

        Returns
        -------
        :class:`~signac.contrib.collection._CollectionSearchResults`
            A result object that iterates over all matching documents.

        Raises
        ------
        ValueError
            In case that the filter argument is invalid.

        """
        return _CollectionSearchResults(self, self._find(filter, limit=limit))

    def find_one(self, filter=None):
        """Return one document that matches the filter or None.

        .. code-block:: python

            doc = collection.find_one(my_filter)
            if doc is None:
                print("No result found for filter", my_filter)
            else:
                print("Doc matching filter:", my_filter, doc)

        Parameters
        ----------
        filter : dict
            The returned document must match the given filter (Default value = None).

        Returns
        -------
        dict
            A matching document or None.

        Raises
        ------
        ValueError
            In case that the filter argument is invalid.

        """
        for doc in self.find(filter, limit=1):
            return doc

    def replace_one(self, filter, replacement, upsert=False):
        """Replace one document that matches the given filter.

        The first document matching the filter will be replaced
        by the given replacement document. If the `upsert` argument
        is True, the replacement will be inserted in case that
        no document matches the filter.

        Parameters
        ----------
        filter : dict
            A document that should be replaced must match this filter.
        replacement : dict
            The replacement document.
        upsert : bool
            If True, insert the replacement document in
            the case that no document matches the filter (Default value = False).

        Returns
        -------
        str
            The id of the replaced (or upserted) documented.

        Raises
        ------
        ValueError
            In case that the filter argument is invalid.

        """
        self._assert_open()
        if len(filter) == 1 and self._primary_key in filter:
            _id = filter[self._primary_key]
            if upsert or _id in self:
                self[_id] = replacement
                return _id
        else:
            for _id in self._find(filter):
                self[_id] = replacement
                return _id
            else:
                if upsert:
                    return self.insert_one(replacement)

    def delete_many(self, filter):
        """Delete all documents that match the filter.

        Parameters
        ----------
        filter : dict
            A document that should be deleted must match this filter.

        """
        to_delete = set(self._find(filter))
        for _id in to_delete:
            del self[_id]

    def delete_one(self, filter):
        """Delete one document that matches the filter.

        Parameters
        ----------
        filter : dict
            The document that should be deleted must match this filter.

        """
        to_delete = set(self._find(filter, limit=1))
        for _id in to_delete:
            del self[_id]

    def _dump(self, text_buffer):
        """Dump collection content serialized to JSON to text-buffer.

        Parameters
        ----------
        text_buffer :
            The file to write the content serialized to JSON (Default value = sys.stdout).

        """
        for doc in self._docs.values():
            text_buffer.write(json.dumps(doc) + "\n")

    def dump(self, file=sys.stdout):
        """Dump the collection in JSON-encoding to file.

        The file argument defaults to `sys.stdout`, which means
        the encoded blob will be printed to screen in case
        that no file argument is provided.

        For example, to dump to a file on disk, one could write:

        .. code-block:: python

            with open('my_collection.txt', 'w') as file:
                collection.dump(file)

        Parameters
        ----------
        file :
            The file to write the encoded blob to (Default value = sys.stdout).

        """
        self._assert_open()
        if self._compresslevel > 0:
            import gzip

            with gzip.GzipFile(
                compresslevel=self._compresslevel, fileobj=file, mode="wb"
            ) as gzipfile:
                text_io = io.TextIOWrapper(gzipfile, encoding="utf-8")
                self._dump(text_io)
                text_io.flush()
        else:
            self._dump(file)

    def to_json(self, file=None):
        """Dump the collection as a JSON file.

        This function returns the JSON-string directly if the
        file argument is None.

        Parameters
        ----------
        file :
            The filename or a file-like object to write the JSON string to (Default value = None).

        Returns
        -------
        JSON
            JSON-string when file argument is not provided.

        """
        json_string = json.dumps(list(self.find()))
        if file is None:
            return json_string
        elif isinstance(file, str):
            with open(file, "w") as json_file:
                json_file.write(json_string)
        else:
            file.write(json_string)

    @classmethod
    def read_json(cls, file=None):
        """Construct an instance of Collection from a JSON file.

        Parameters
        ----------
        file :
            The json file to read, provided as either a filename or a
            file-like object (Default value = None).

        Returns
        -------
        :class:`~Collection`
            A Collection containing the JSON file

        """
        if isinstance(file, str):
            with open(file) as json_file:
                json_data = json.load(json_file)
        else:
            json_data = json.load(file)
        return Collection(json_data)

    @classmethod
    def _open(cls, file, compresslevel=0):
        """Open a collection associated with a file on disk.

        Parameters
        ----------
        file :
            The file to read the documents from or create the file if it does not exist.
        compresslevel : int
            The level of compression to use. Any positive value
            implies compression and is used by the underlying gzip implementation.
            (Default value = 0)

        Returns
        -------
        :class:`~Collection`
            An instance of :class:`~Collection`.

        """
        try:
            if compresslevel > 0:
                import gzip

                with gzip.GzipFile(fileobj=file, mode="rb") as gzipfile:
                    text_io = io.TextIOWrapper(gzipfile, encoding="utf-8")
                    collection = cls(docs=(json.loads(line) for line in text_io))
                    text_io.detach()
            else:
                collection = cls(docs=(json.loads(line) for line in file))
        except (OSError, io.UnsupportedOperation) as error:
            if str(error) in ("not readable", "read"):
                collection = cls()
            else:
                raise error
        except ValueError as error:
            file.close()
            if hasattr(file, "name"):
                raise JSONParseError(
                    f"Error while trying to parse file '{file.name}': {error}."
                )
            else:
                raise JSONParseError(f"Error while trying to parse '{file}': {error}.")
        except AttributeError as e:
            # This error occurs in python27 and has been evaluated as being
            # fine to accept in this manner
            if str(e) == "'GzipFile' object has no attribute 'extrastart'":
                collection = cls()
            else:
                raise AttributeError(e)
        collection._file = file
        collection._compresslevel = compresslevel
        collection._requires_flush = False  # not needed after initial read
        return collection

    @classmethod
    def open(cls, filename, mode=None, compresslevel=None):
        """Open a collection associated with a file on disk.

        Using this factory method will return a collection that is
        associated with a collection file on disk. For example:

        .. code-block:: python

            with Collection.open('collection.txt') as collection:
                for doc in collection:
                    print(doc)

        will read all documents from the `collection.txt` file or create
        the file if it does not exist yet.

        Modifications to the file will be written to the file when the
        :meth:`~Collection.flush` method is called or the collection is
        explicitly closed by calling the :meth:`Collection.close` method or
        implicitly by leaving the `with`-clause:

        .. code-block:: python

            with Collection.open('collection.txt') as collection:
                collection.update(my_docs)
            # All changes to the collection have been written to collection.txt.

        The open-modes work as expected, so for example to open a collection
        file in *read-only* mode, use ``Collection.open('collection.txt', 'r')``.

        Opening a gzip (`*.gz`) file also works as expected. Because gzip does not
        support a combined read and write mode, `mode=*+` is not available. Be
        sure to open the file in read, write, or append mode as required. Due to
        the manner in which gzip works, opening a file in `mode=wt` will
        effectively erase the current file, so take care using `mode=wt`.

        Parameters
        ----------
        filename : str
            Name of file to read the documents from or create the file if it does not exist.
        mode : str
            Open the file with mode (Default value = None).
        compresslevel : int
            The level of compression to use. Any positive value
            implies compression and is used by the underlying gzip implementation.
            (Default value = None)

        Returns
        -------
        :class:`~Collection`
            An instance of :class:`~Collection`.

        Raises
        ------
        RuntimeError
            File open-mode is not None for in-memory collection or
            compressed collections are not opened in binary mode.

        """
        if compresslevel is None:
            compresslevel = 9 if filename.endswith(".gz") else 0

        logger.debug(f"Open collection '{filename}'.")
        if filename == ":memory:":
            if mode is not None:
                raise RuntimeError(
                    "File open-mode must be None for in-memory collection."
                )
            return cls(compresslevel=compresslevel)  # That's the default open mode.
        else:
            # Set default mode
            if mode is None:
                mode = "ab+"

            file = open(filename, mode)
            file.seek(0)

            if "b" in mode:
                if compresslevel > 0:
                    return cls._open(file, compresslevel=compresslevel)
                else:
                    return cls._open(io.TextIOWrapper(file, encoding="utf-8"))
            elif compresslevel > 0:
                raise RuntimeError(
                    "Compressed collections must be opened in binary mode, for example: 'ab+'."
                )
            else:
                return cls._open(file)

    def flush(self):
        """Write all changes to the associated file.

        If the collection instance is associated with a file-object,
        calling the :meth:`~Collection.flush` method will write all changes
        to this file.

        This method is also called when the collection is explicitly or
        implicitly closed.

        """
        self._assert_open()
        if self._requires_flush:
            if self._file is None:
                logger.debug("Flushed collection.")
            else:
                logger.debug(f"Flush collection to file '{self._file}'.")
                try:
                    self._file.truncate(0)
                except ValueError as error:
                    raise error
                else:
                    self.dump(self._file)
                    self._file.flush()
            self._requires_flush = False
        else:
            logger.debug("Flushed collection (no changes).")

    def close(self):
        """Close this collection instance.

        In case that the collection is associated with a file-object,
        all changes are flushed to the file and the file is closed.

        It is not possible to re-open the same collection instance
        after closing it.

        """
        if self._file is not None:
            try:
                self.flush()
            finally:
                self._file.close()
                self._indexes.clear()
                self._docs = None
                self._file = None

    def __enter__(self):
        return self

    def __exit__(self, t, v, tb):
        self.close()

    def main(self):
        """Start a command line interface for this Collection.

        Use this function to interact with this instance of Collection
        on the command line. For example, executing the following script:

        .. code-block:: python

            # find.py
            with Collection.open('my_collection.txt') as c:
                c.main()

        will enable us to search for documents on the command line like this:

        .. code-block:: bash

            $ python find.py '{"age": 32}'
            {"name": "John", "age": 32}
            {"name": "Kevin", "age": 32}

        Raises
        ------
        ValueError
            When both `--id` or `--indent` are selected.

        """
        parser = argparse.ArgumentParser(
            "Command line interface for instances of Collection."
        )
        parser.add_argument(
            "filter",
            nargs="*",
            help="The search filter provided in JSON encoding. "
            "Leave empty to return all documents.",
        )
        parser.add_argument(
            "-l",
            "--limit",
            type=int,
            default=0,
            help="Limit the number of search results that are "
            "maximally returned. A value of 0 (the default) "
            "means no limit.",
        )
        parser.add_argument(
            "--id",
            dest="_id",
            action="store_true",
            help="Print a document's primary key instead of the whole document.",
        )
        parser.add_argument(
            "-i",
            "--indent",
            action="store_true",
            help="Print results in indented format.",
        )
        args = parser.parse_args()
        if args._id and args.indent:
            raise ValueError("Select either `--id` or `--indent`, not both.")
        f = parse_filter_arg(args.filter)
        for doc in self.find(f, limit=args.limit):
            if args._id:
                print(doc[self._primary_key])
            else:
                if args.indent:
                    print(json.dumps(doc, indent=2))
                else:
                    print(json.dumps(doc))



# Copyright (c) 2017 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
import logging
import subprocess
from os.path import expanduser

import pymongo
from deprecation import deprecated

from ..version import __version__
from .errors import ConfigError

PYMONGO_2 = pymongo.version_tuple[0] == 2

logger = logging.getLogger(__name__)

AUTH_NONE = "none"
AUTH_SCRAM_SHA_1 = "SCRAM-SHA-1"
AUTH_SSL = "SSL"
AUTH_SSL_x509 = "SSL-x509"

"""
THIS MODULE IS DEPRECATED!
"""


@deprecated(
    deprecated_in="1.3",
    removed_in="2.0",
    current_version=__version__,
    details="The connection module is deprecated.",
)
def get_subject_from_certificate(fn_certificate):  # pragma no cover
    try:
        cert_txt = subprocess.check_output(
            [
                "openssl",
                "x509",
                "-in",
                fn_certificate,
                "-inform",
                "PEM",
                "-subject",
                "-nameopt",
                "RFC2253",
            ]
        ).decode()
    except subprocess.CalledProcessError:
        msg = "Unable to retrieve subject from certificate '{}'."
        raise RuntimeError(msg.format(fn_certificate))
    else:
        lines = cert_txt.split("\n")
        assert lines[0].startswith("subject=")
        return lines[0][len("subject=") :].strip()


@deprecated(
    deprecated_in="1.3",
    removed_in="2.0",
    current_version=__version__,
    details="The connection module is deprecated.",
)
def raise_unsupported_auth_mechanism(mechanism):
    msg = "Auth mechanism '{}' not supported."
    raise ValueError(msg.format(mechanism))


@deprecated(
    deprecated_in="1.3",
    removed_in="2.0",
    current_version=__version__,
    details="The connection module is deprecated.",
)
class DBClientConnector:
    def __init__(self, host_config, **kwargs):
        self._config = host_config
        self._client = None
        self._kwargs = kwargs

    @property
    def client(self):
        if self._client is None:
            raise RuntimeError("Client not connected.")
        else:
            return self._client

    @property
    def host(self):
        return self._config["url"]

    @property
    def config(self):
        return dict(self._config)

    def _config_get(self, key, default=None):
        return self._config.get(key, default)

    def _config_get_required(self, key):
        try:
            return self._config[key]
        except KeyError as e:
            raise ConfigError(f"Missing required key '{e}'.")

    def _connect_pymongo3(self, host):
        logger.debug("Connecting with pymongo3.")
        forwarded_parameters = (
            "socketTimeoutMS",
            "connectTimeoutMS",
            "serverSelectionTimeoutMS",
            "w",
            "wtimeout",
            "replicaSet",
        )
        parameters = self._kwargs
        for parameter in forwarded_parameters:
            if parameter in self._config:
                parameters[parameter] = self._config_get(parameter)

        auth_mechanism = self._config_get("auth_mechanism", AUTH_NONE)
        if auth_mechanism in (AUTH_NONE, AUTH_SCRAM_SHA_1):
            client = pymongo.MongoClient(
                host,
                read_preference=getattr(
                    pymongo.read_preferences.ReadPreference,
                    self._config_get("read_preference", "PRIMARY"),
                ),
                **parameters,
            )
        else:
            raise_unsupported_auth_mechanism(auth_mechanism)
        self._client = client

    def connect(self, host=None):
        if host is None:
            host = self._config_get_required("url")
        logger.debug(
            "Connecting to host '{host}'.".format(host=self._config_get_required("url"))
        )

        if PYMONGO_2:
            raise RuntimeError("pymongo version 2.x is no longer supported.")
        else:
            self._connect_pymongo3(host)

    def authenticate(self):
        auth_mechanism = self._config_get("auth_mechanism", AUTH_NONE)
        logger.debug(f"Authenticating: mechanism={auth_mechanism}")
        if auth_mechanism == AUTH_SCRAM_SHA_1:
            db_auth = self.client[self._config.get("db_auth", "admin")]
            username = self._config_get_required("username")
            msg = "Authenticating user '{user}' with database '{db}'."
            logger.debug(msg.format(user=username, db=db_auth))
            db_auth.authenticate(
                username,
                self._config_get_required("password"),
                mechanism=AUTH_SCRAM_SHA_1,
            )
        elif auth_mechanism in (AUTH_SSL, AUTH_SSL_x509):  # pragma no cover
            certificate_subject = get_subject_from_certificate(
                expanduser(self._config_get_required("ssl_certfile"))
            )
            logger.debug(f"Authenticating: user={certificate_subject}")
            db_external = self.client["$external"]
            db_external.authenticate(certificate_subject, mechanism="MONGODB-X509")
        elif auth_mechanism == AUTH_NONE:
            pass
        else:
            raise_unsupported_auth_mechanism(auth_mechanism)

    def logout(self):
        auth_mechanism = self._config_get("auth_mechanism", AUTH_NONE)
        if auth_mechanism == AUTH_SCRAM_SHA_1:
            db_auth = self.client["admin"]
            db_auth.logout()
        elif auth_mechanism in (AUTH_SSL, AUTH_SSL_x509):  # pragma no cover
            db_external = self.client["$external"]
            db_external.logout()
        elif auth_mechanism == AUTH_NONE:
            pass
        else:
            raise_unsupported_auth_mechanism(auth_mechanism)



# Copyright (c) 2017 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
import getpass
import logging
import warnings

from deprecation import deprecated

from ..core import json
from ..version import __version__
from .config import load_config
from .connection import DBClientConnector
from .crypt import SimpleKeyring, get_crypt_context, get_keyring
from .errors import AuthenticationError, ConfigError

logger = logging.getLogger(__name__)


SESSION_PASSWORD_HASH_CACHE = SimpleKeyring()
SESSION_USERNAME_CACHE = {}  # type: ignore

"""
THIS MODULE IS DEPRECATED!
"""


@deprecated(
    deprecated_in="1.3",
    removed_in="2.0",
    current_version=__version__,
    details="The host module is deprecated.",
)
def get_default_host(config=None):
    if config is None:
        config = load_config()
    try:
        return config["General"]["default_host"]
    except KeyError:
        try:
            return config["hosts"].keys()[0]
        except (KeyError, IndexError):
            raise ConfigError("No hosts specified.")


def _get_host_config(hostname, config):
    try:
        return config["hosts"][hostname]
    except KeyError:
        raise ConfigError(f"Host '{hostname}' not configured.")


@deprecated(
    deprecated_in="1.3",
    removed_in="2.0",
    current_version=__version__,
    details="The host module is deprecated.",
)
def get_host_config(hostname=None, config=None):
    if config is None:
        config = load_config()
    if hostname is None:
        hostname = get_default_host(config)
    return _get_host_config(hostname, config)


def _host_id(hostcfg):
    return json.dumps(hostcfg, sort_keys=True)


@deprecated(
    deprecated_in="1.3",
    removed_in="2.0",
    current_version=__version__,
    details="The host module is deprecated.",
)
def make_uri(hostcfg):
    ret = hostcfg["url"]
    if ret.startswith("mongodb://"):
        ret = ret[len("mongodb://") :]
    if "username" in hostcfg:
        assert "@" not in ret
        ret = hostcfg["username"] + "@" + ret
    return "mongodb://" + ret


def _request_credentials(hostcfg):
    pwcfg = hostcfg.get("password_config")
    pw = getpass.getpass("Enter password for {}: ".format(make_uri(hostcfg)))
    if pwcfg and "salt" in pwcfg and "rounds" in pwcfg:
        logger.debug("Using password configuration for hashing.")
        return get_crypt_context().encrypt(pw, **pwcfg)
    else:
        return pw


def _get_config_credentials(hostcfg):
    return hostcfg.get("password")


def _get_keyring_credentials(hostcfg):
    pwcfg = hostcfg.get("password_config")
    kr = get_keyring()
    if kr is None:
        if pwcfg and "keyring" in pwcfg:
            warnings.warn("Password stored in keyring, but keyring is not available!")
    elif pwcfg and "keyring" in pwcfg:
        return kr.get_password("signac", pwcfg["keyring"])
    else:
        return kr.get_password("signac", make_uri(hostcfg))


def _get_cached_credentials(hostcfg, default):
    hostcfg_id = json.dumps(hostcfg, sort_keys=True)
    if hostcfg_id in SESSION_PASSWORD_HASH_CACHE:
        logger.debug("Loading credentials from cache.")
        return SESSION_PASSWORD_HASH_CACHE[hostcfg_id]
    else:
        return SESSION_PASSWORD_HASH_CACHE.setdefault(hostcfg_id, default())


def _get_stored_credentials(hostcfg):
    def default():
        pw = _get_keyring_credentials(hostcfg)
        if pw is None:
            pw = _get_config_credentials(hostcfg)
        return pw

    return _get_cached_credentials(hostcfg, default)


def _get_credentials(hostcfg):
    def default():
        pw = _get_config_credentials(hostcfg)
        if pw is None:
            pw = _get_keyring_credentials(hostcfg)
            if pw is None:
                pw = _request_credentials(hostcfg)
        return pw

    return _get_cached_credentials(hostcfg, default)


@deprecated(
    deprecated_in="1.3",
    removed_in="2.0",
    current_version=__version__,
    details="The host module is deprecated.",
)
def get_credentials(hostcfg, ask=True):
    if ask:
        return _get_credentials(hostcfg)
    else:
        return _get_stored_credentials(hostcfg)


def _input(prompt, default=""):
    try:
        value = input(prompt)
    except SyntaxError:
        return default
    if value:
        return value
    else:
        return default


@deprecated(
    deprecated_in="1.3",
    removed_in="2.0",
    current_version=__version__,
    details="The host module is deprecated.",
)
def check_credentials(hostcfg):
    from pymongo.uri_parser import parse_uri

    auth_m = hostcfg.get("auth_mechanism", "none")
    if auth_m == "SCRAM-SHA-1":
        uri = hostcfg["url"]
        if isinstance(uri, list):
            uri = ",".join(uri)
        if "username" not in hostcfg and not parse_uri(uri)["username"]:
            username = SESSION_USERNAME_CACHE.get(_host_id(hostcfg))
            if username:
                hostcfg["username"] = username
            else:
                SESSION_USERNAME_CACHE[_host_id(hostcfg)] = hostcfg[
                    "username"
                ] = _input(f"Username ({getpass.getuser()}): ", getpass.getuser())
        if "password" not in hostcfg and not parse_uri(uri)["password"]:
            hostcfg["password"] = get_credentials(hostcfg)
    return hostcfg


@deprecated(
    deprecated_in="1.3",
    removed_in="2.0",
    current_version=__version__,
    details="The host module is deprecated.",
)
def get_connector(hostcfg, **kwargs):
    return DBClientConnector(hostcfg, **kwargs)


@deprecated(
    deprecated_in="1.3",
    removed_in="2.0",
    current_version=__version__,
    details="The host module is deprecated.",
)
def get_client(hostcfg, **kwargs):
    connector = get_connector(hostcfg, **kwargs)
    connector.connect()
    connector.authenticate()
    return connector.client


@deprecated(
    deprecated_in="1.3",
    removed_in="2.0",
    current_version=__version__,
    details="The host module is deprecated.",
)
def get_database(name, hostname=None, config=None, **kwargs):
    if hostname is None:
        hostname = get_default_host(config)
    if config is None:
        config = load_config()
    hostcfg = check_credentials(get_host_config(hostname, config))
    logger.debug(
        "Connecting with host config: {}".format(
            {k: "***" if "password" in k else v for k, v in hostcfg.items()}
        )
    )
    try:
        client = get_client(hostcfg, **kwargs)
    except Exception as error:
        if "Authentication failed" in str(error):
            raise AuthenticationError(hostname)
        raise
    return client[name]



# Copyright (c) 2018 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.


"""Module for signac deprecation warnings."""


class SignacDeprecationWarning(UserWarning):
    """Indicates the deprecation of a signac feature, API or behavior.

    This class indicates a user-relevant deprecation and is therefore
    a UserWarning, not a DeprecationWarning which is hidden by default.
    """

    pass


__all__ = ["SignacDeprecationWarning"]



# Copyright (c) 2017 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""Validate config schema."""

import logging

from .configobj.validate import Validator, VdtValueError

logger = logging.getLogger(__name__)


def mongodb_uri(value, *args, **kwargs):
    """Return a MongoDB URI."""
    if isinstance(value, list):
        value = ",".join(value)
    if not value.startswith("mongodb://"):
        value = "mongodb://" + value
    try:
        import pymongo
    except ImportError:
        logger.debug("Install pymongo to validate database configurations!")
    else:
        try:
            pymongo.uri_parser.parse_uri(value)
        except pymongo.errors.InvalidURI:
            raise VdtValueError(value)
    return value


def password(value, *args, **kwargs):  # noqa: D103
    return value


def get_validator():  # noqa: D103
    return Validator({"mongodb_uri": mongodb_uri, "password": password})


cfg = """
project = string()
workspace_dir = string(default='workspace')
schema_version = string(default='1')

[General]
default_host = string(default=None)

[hosts]
[[__many__]]
url = mongodb_uri(default='localhost')
auth_mechanism = option('none', 'SCRAM-SHA-1', default='none')
username = string()
password = password()
db_auth = string(default='admin')
[[[password_config]]]
salt = string()
rounds = integer()
"""



# Copyright (c) 2017 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""Errors raised by signac."""

# The subpackage error modules (e.g. signac.core.errors) are used to bundle
# exceptions that are relevant beyond a single module. This top-level errors
# module is used to expose user-facing exception classes.

from .common.errors import AuthenticationError, ConfigError, ExportError, FetchError
from .contrib.errors import (
    DestinationExistsError,
    IncompatibleSchemaVersion,
    JobsCorruptedError,
)
from .core.errors import Error
from .core.jsondict import BufferedFileError, BufferException


class SyncConflict(Error, RuntimeError):
    """Raised when a synchronization operation fails."""

    pass


class FileSyncConflict(SyncConflict):
    """Raised when a synchronization operation fails due to a file conflict."""

    def __init__(self, filename):
        self.filename = filename
        "The filename of the file that caused the conflict."

    def __str__(self):
        return f"The file with filename '{self.filename}' caused a conflict."


class DocumentSyncConflict(SyncConflict):
    """Raised when a synchronization operation fails due to a document conflict."""

    def __init__(self, keys):
        self.keys = keys
        "The keys that caused the conflict."

    def __str__(self):
        return "The following keys caused a conflict: {}".format(", ".join(self.keys))


class SchemaSyncConflict(SyncConflict):
    """Raised when a synchronization operation fails due to schema differences."""

    def __init__(self, schema_src, schema_dst):
        self.schema_src = schema_src
        self.schema_dst = schema_dst

    def __str__(self):
        return "The synchronization failed, because of a schema conflict."


class InvalidKeyError(ValueError):
    """Raised when a user uses a non-conforming key."""


class KeyTypeError(TypeError):
    """Raised when a user uses a key of invalid type."""


__all__ = [
    "Error",
    "BufferException",
    "BufferedFileError",
    "ConfigError",
    "AuthenticationError",
    "ExportError",
    "FetchError",
    "DestinationExistsError",
    "JobsCorruptedError",
    "IncompatibleSchemaVersion",
    "SyncConflict",
    "FileSyncConflict",
    "DocumentSyncConflict",
    "SchemaSyncConflict",
    "InvalidKeyError",
]



# Copyright (c) 2020 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
import os

from setuptools import find_packages, setup

requirements = [
    # Deprecation management
    "deprecation>=2",
    # Platform-independent file locking
    "filelock~=3.0",
    # Used for version parsing and comparison
    "packaging>=15.0",
    # Progress bars
    "tqdm>=4.10.0",
]

description = "Manage large and heterogeneous data spaces on the file system."

try:
    this_path = os.path.dirname(os.path.abspath(__file__))
    fn_readme = os.path.join(this_path, "README.md")
    with open(fn_readme) as fh:
        long_description = fh.read()
except OSError:
    long_description = description

setup(
    name="signac",
    version="1.5.1",
    packages=find_packages(),
    zip_safe=True,
    maintainer="signac Developers",
    maintainer_email="signac-support@umich.edu",
    author="Carl Simon Adorf et al.",
    author_email="csadorf@umich.edu",
    description=description,
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://signac.io",
    download_url="https://pypi.org/project/signac/",
    keywords="simulation database index collaboration workflow",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: BSD License",
        "Topic :: Database",
        "Topic :: Scientific/Engineering :: Physics",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
    ],
    project_urls={
        "Homepage": "https://signac.io",
        "Documentation": "https://docs.signac.io",
        "Source Code": "https://github.com/glotzerlab/signac",
        "Issue Tracker": "https://github.com/glotzerlab/signac/issues",
    },
    install_requires=requirements,
    # Supported versions are determined according to NEP 29.
    # https://numpy.org/neps/nep-0029-deprecation_policy.html
    python_requires=">=3.6, <4",
    extras_require={"db": ["pymongo>=3.0"], "mpi": ["mpi4py"], "h5": ["h5py"]},
    entry_points={"console_scripts": ["signac = signac.__main__:main"]},
)
